import requests
import pandas as pd

###################### ส่วนที่ใช้ Library ภายใน ###################
try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    #import config as con 
    from starfishX import config as con
except:
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""  
    
    
def getPriceAndRatioInSector(symbol,date=''):
  #resp = requests.get('https://www.settrade.com/api/set/stock/KCE/rank-in-sector?type=percentChange&date=29/04/2022')
  if(date!=''):
    date = '?type=percentChange&date='+date
    resp = requests.get('https://www.settrade.com/api/set/stock/'+symbol+'/rank-in-sector'+date)
  else:
    resp = requests.get('https://www.settrade.com/api/set/stock/'+symbol+'/rank-in-sector')
  txt = resp.json()
  df = pd.json_normalize(txt['stockTradings'])
  df['date'] = pd.to_datetime(df['date'].str[:10])
  df = df[['date','symbol','close','dividendYield','marketCap','pe','pbv']]

  return df


def listStockInSector(symbol):
  resp = requests.get('https://www.settrade.com/api/set/stock/'+symbol+'/rank-in-sector')
  txt = resp.json()
  df = pd.json_normalize(txt['stockTradings'])
  df['date'] = pd.to_datetime(df['date'].str[:10])
  df[['date','symbol','close','dividendYield','marketCap','pe','pbv']]
 
  return df[['symbol']].values.reshape(1,-1).tolist()[0]

def getRatioAndBalanceSheet(symbol):
    url = 'https://www.settrade.com/api/set/stock/'+symbol+'/company-highlight/financial-data'
    resp = requests.get(url)
    txt = resp.json()
    df = pd.json_normalize(txt)
    df['date'] = df['year'].astype(str)+df['quarter']
    #df
    #df = df[df['year']==2021]
    col = df['date']
    df = df[['date','totalAsset','totalLiability','equity','grossProfitMargin',
        'netProfitMargin','roe','roa','deRatio','eps']].T
    df.columns = col
    df = df.drop('date')
    return df

def getFundamentalInSector(symbol,year):
    '''
    symbol : str ชื่อย่อของหุ้น
    year : ปีคศ. เช่น 2021,2022 เป็นต้น
    
    date_x คือวันที่ของราคาหุ้น
    date_y คือวันที่ปิดงบ
    '''
    df1 = getPriceAndRatioInSector(symbol)
     
    s = 0
    for symbol in df1.symbol:
        df = getRatioAndBalanceSheet(symbol)
        k = ''
        for i in df.columns:
          if(str(year) in i):
             k = (i)
        if(k==''):
           continue #ถ้าไม่มีปีก็ข้ามไปเลย

        k = df[[k]].T
        k = k.reset_index()
        k['symbol'] = symbol

        if(s==0):
          df2 = k
          s=1
        else:
          df2 = df2.append(k)
        
    df = pd.merge(df1,df2,on='symbol')
    return df


#scan Peter Lynch

import requests
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np

import starfishX as sx


import re
import ssl
from urllib import request, parse

import os
from selenium import webdriver
from selenium.webdriver.support.ui import Select
import datetime

from IPython.display import HTML

#สำหรับหาความแข็งแกร่งของแนวโน้ม
from scipy import stats


pd.options.display.float_format = '{:,.2f}'.format
#pd.set_option('precision', 2)


def getPE(symbol):
 '''
 หา PE ของหุ้น ,อุตสาหกรรม และตลาด
 parameter : รายชื่อหุ้น symbol

 return  
 - PE บริษัท
 - PE อุตสาหกรรม
 - PE ของตลาด (SET)
 '''  
 symbol = symbol.replace("&","%26") 
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 url = "https://www.set.or.th/set/factsheet.do?symbol="+symbol+"&ssoPageId=3&language=th&country=TH"
 r = requests.post(url,headers=headers, timeout=5)
 soup = BeautifulSoup(r.content, "lxml")  

 ct = soup.find_all("table",{"class":"table-factsheet-padding0"})  

 if(len(ct)==0):
    return 0

 #### หา index #### content อยู่ช่องไหน
 i_index = 0
 for i in ct:
  if("เปรียบเทียบหมวดอุตสาหกรรม" in i.text):
      m_index = i_index
      break
  i_index+=1

 #######
 tr = ct[m_index].find_all("tr")
 for i in tr:
  td = i.find_all("td")
  if("P/E" in td[0].text):
      try:
       PE_self  = float(td[1].text)
       PE_indus = float(td[2].text)
       PE_SET   = float(td[3].text)  
       pe_df = pd.DataFrame({"PE":PE_self,"PE_INDS":PE_indus,"PE_SET":PE_SET},index=[symbol.upper()])  
       return pe_df
      except: 
       return 0  


def getEPSGrowth(symbol,backwardYear=5):
    '''
    หาอัตราการเติบโตของ EPS ย้อนหลังกี่ปี ถ้าไม่ระบุจะใช้ 5 ปี
    symbol หุ้น
    backward_year : จำนวนปีย้อนหลังนับจากปีปัจจุบัน
    
    return g และ log
    '''
    eps_g = sx.morningstarGetfn(symbol,sx.MStar.EarningsPerShare)
    
    eps_now = eps_g.iloc[-1].values[0]
    eps_backward = eps_g.iloc[-backwardYear].values[0]

    k = eps_now/eps_backward 
    if(k>=0):
      g = (k**(1/backwardYear))-1
      return g,eps_g
    else:
      return np.NaN,eps_g 

def getListedShare(symbol,andPrice=False):
 '''
 หา จำนวนหุ้นของบริษัทจดทะเบียน(ล้านหุ้น)
 symbol : str สัญลักษณ์ชื่อของหุ้น
 andPrice : boolean ถ้าเป็น True จะคืนค่าราคาแถมมาด้วย
 return  
 - float (ล้านหุ้น) , ราคาหุ้นล่าสุด
 '''  
 symbol = symbol.replace("&","%26") 
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 url = "https://www.set.or.th/set/factsheet.do?symbol="+symbol+"&ssoPageId=3&language=th&country=TH"
 r = requests.post(url,headers=headers, timeout=5)
 soup = BeautifulSoup(r.content, "lxml")  

 ct = soup.find_all("table",{"class":"table-factsheet-padding0"})  

 if(len(ct)==0):
    return 0

 #### หา index #### content อยู่ช่องไหน
 i_index = 0
 for i in ct:
  if("จำนวนหุ้นจดทะเบียน (ล้านหุ้น)" in i.text):
      m_index = i_index
      break
  i_index+=1

 #######

 tr = ct[m_index].find_all("tr")
 for i in tr:
  td = i.find_all("td") 
  if("จำนวนหุ้นจดทะเบียน" in td[0].text):
      listedShare  = td[1].text
      try:
       listedShare = float(listedShare.replace(",",""))
      except:
       return 0
   
  if("ราคา (บาท/หุ้น)" in td[0].text):
      priceNow  = td[1].text
      try:
       priceNow = float(priceNow.replace(",",""))
      except:
       return 0
       
      if(andPrice==True):
         #ls_df = pd.DataFrame({"listedShare":listedShare,"price":priceNow},index=[symbol]) 
         return priceNow
      else:
         #ls_df = pd.DataFrame({"listedShare":listedShare},index=[symbol])
         return listedShare
      #return ls_df  

def getInterestBearingDebt(symbol, keywordlist=['เงินกู้', 'ตราสารหนี้', 'หุ้นกู้', 'เงินเบิกเกินบัญชี','หนี้สินระยะยาว - สุทธิจากส่วนที่ถึงกำหนดชำระภายในหนึ่งปี','ส่วนของหนี้สินระยะยาวที่ถึงกำหนดชำระภายในหนึ่งปี'], viewlog=False):
  TL,IBD,Log = sx.InterestBearingDebt(symbol, keywordlist, viewlog)
  return TL,IBD,Log

def getCash(symbol):
   bs = getBalanceSheet(symbol)
   cash = bs[bs.columns[0]][bs.index=="เงินสด"]
   cash = float(cash.values[0].replace(",",""))
   return cash

def getEquity(symbol):
   bs = getBalanceSheet(symbol)
   eq = bs[bs.columns[0]][bs.index=="ส่วนของผู้ถือหุ้นบริษัทใหญ่"]
   eq = float(eq.values[0].replace(",",""))
   return eq

def getBalanceSheet(symbol):
   bs = sx.getBalanceSheet(symbol)
   return bs

def getPrice(symbol):
   price = getListedShare(symbol,andPrice=True)
   #price = price_df["price"].values[0]
   return price

def reportHelper(*condition,symbol,backwardYear=5):
  rp = pd.DataFrame()  
  for i in condition:
    if(i.value=="epsGrowth"):
        g,log = (getEPSGrowth(symbol,backwardYear))
        rp["Eps Growth"] = [g]
    
    
    if(i.value=="PE/G"):
        pe_df  = (getPE(symbol))
        pe_self  = pe_df["PE"].values[0]  
        pe_indus = pe_df["PE_INDS"].values[0] 
        pe_set = pe_df["PE_SET"].values[0]  
        rp["P/E"] = [pe_self]
        rp["P/E Indus"] = [pe_indus]
        rp["P/E SET"] = [pe_set]

        try:
           peg = pe_self/g
           rp["PE/G"] = [peg]
        except:
           g,log = (getEPSGrowth(symbol,backwardYear))
           peg = pe_self/(g*100)
           rp["Eps Growth("+str(backwardYear)+"Years)"] = [g]
           rp["PE/G"] = [peg]

    if(i.value=="PE"):
        pe_df  = (getPE(symbol))
        pe_self  = pe_df["PE"].values[0]  
        pe_indus = pe_df["PE_INDS"].values[0] 
        pe_set = pe_df["PE_SET"].values[0]  
        rp["P/E"] = [pe_self]
        rp["P/E Indus"] = [pe_indus]
        rp["P/E SET"] = [pe_set]

    if(i.value=="cash"):
        bs = getBalanceSheet(symbol)
        cash = bs[bs.columns[0]][bs.index=="เงินสด"]
        cash = float(cash.values[0].replace(",",""))
        rp["Cash"] = [cash]
        
    if(i.value=="listedShare"):
        share = getListedShare(symbol,andPrice=False)
        #share = share_df["listedShare"].values[0]
        rp["ListedShare"] = [share] 
        
    if(i.value=="price"):
        price = getPrice(symbol)
        rp["Price"] = [price]    
        
    if(i.value=="interestBearingDebt"):
        TL,IBD,Log = getInterestBearingDebt(symbol,viewlog=False)
        rp["Total Liabilities"] = [TL]
        rp["Interest Bearing Debt"] = [IBD]

    if(i.value=="equity"):
        equity = getEquity(symbol)
        rp["Equity"] = [equity]    
    
    if(i.value=="de"):
        try:
           DE = IBD/equity
           rp["DE"] = [DE]
        except:
           TL,IBD,Log = getInterestBearingDebt(symbol,viewlog=False)
           rp["Total Liabilities"] = [TL]
           rp["Interest Bearing Debt"] = [IBD]

           equity = getEquity(symbol)
           rp["Equity"] = [equity] 
           DE = IBD/equity
           rp["D/E"] = [DE]   

    if(i.value=="stockRepurchases"):
        stockRepurchases = getStockRepurchasesCheck(symbol)
        rp["Stock Repurchases"] = [stockRepurchases] 

    if(i.value=="cashPerShare"):
        try:
           cashPerShare = (cash - IBD)/ share
           rp["Cash/Share"] = [cashPerShare]
        except:   
           bs = getBalanceSheet(symbol)
           cash = bs[bs.columns[0]][bs.index=="เงินสด"]
           cash = float(cash.values[0].replace(",",""))
           rp["Cash"] = [cash]  

           TL,IBD,Log = getInterestBearingDebt(symbol,viewlog=False)
           rp["Total Liabilities"] = [TL]
           rp["Interest Bearing Debt"] = [IBD]

           share = getListedShare(symbol,andPrice=False)
           #share = share_df["listedShare"].values[0]
           rp["ListedShare"] = [share]

           cashPerShare = (cash - IBD)/ share
           rp["Cash/Share"] = [cashPerShare]
    
  ###end for
  rp = rp.T
  rp.columns = [symbol.upper()] 
  return rp.round(2)



def getStockRepurchasesCheck(symbol,action="history"): 
 '''
 action : str จะมีสองค่าคือ "history" และ "today"
   - ถ้า today จะตรวจสอบวันล่าสุด
   - ถ้า history จะตรวจสอบประมาณ 2-3 เดือนย้อนหลัง
 '''  
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 symbolMofi = symbol.replace("&","%26")  
 r = requests.get("https://www.set.or.th/set/companynews.do?symbol="+symbolMofi+"&ssoPageId=8&language=th&country=TH",headers=headers, timeout=5)
 soup = BeautifulSoup(r.content, "lxml")  
 ct = soup.find_all("table",{"class":"table table-hover table-info-wrap"})  
 p = 0
 
 if(action=="today"):
  if("ซื้อหุ้นคืน" in ct[0].text):
    p = 1
 else:
   for i in ct: 
    if("ซื้อหุ้นคืน" in i.text):
      p = 1
    
 if(p==1):
    return True
 else:
    return False


def getDividends(symbol):
  return sx.morningstarGetfn(symbol,sx.MStar.Dividends)

def getPayoutRatio(symbol):
  return sx.morningstarGetfn(symbol,sx.MStar.PayoutRatio)





import re
import pandas as pd
import ssl
from urllib import request, parse
from bs4 import BeautifulSoup

import os
from selenium import webdriver
from selenium.webdriver.support.ui import Select
import datetime

def getProfitAbility(symbol): 
 symbol = symbol.replace("&","%26")    
 symbol = symbol.upper()   
 os.environ['MOZ_HEADLESS'] = '1'  
 driver = webdriver.Firefox()  #อาจต้องลง 
 driver.get("https://www.wsj.com/market-data/quotes/TH/XBKK/"+symbol+"/financials")
 
 #########
    
 lb = []
 data = []
 for i in driver.find_elements_by_class_name("cr_sub_profitability"): 
    #print(i.text+"_")
    for j in i.find_elements_by_class_name("data_lbl"):
     lb.append(j.text)
    for j in i.find_elements_by_class_name("data_data"):
     t = j.text.replace("+","").replace("-","")
     data.append(float(t))
        
 #########
 return pd.DataFrame({symbol:data},index=lb)


#pehistory.py

import requests
from bs4 import BeautifulSoup
import re
import json
import numpy as np
import pandas as pd

def getPEHistory(symbol):
 symboltmp = symbol.replace("&","%26")
 url = "http://financials.morningstar.com/valuate/valuation-history.action?&t=XBKK:"+symboltmp+"&region=tha&culture=en-US&cur=&type=price-earnings"
    
 #symboltmp = symbol.replace("&","%26") 
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 #url = "https://www.set.or.th/set/factsheet.do?symbol="+symboltmp+"&ssoPageId=3&language=th&country=TH"
 r = requests.post(url,headers=headers, timeout=5)
 soup = BeautifulSoup(r.content, "lxml") 

 #######
 regex = "ticks.*tickFormatter"
 p = re.findall(regex,soup.script.text,re.DOTALL)
 p = p[0].replace("ticks:","")
 p = p.replace("\n","")
 p = p.replace("\t","")
 p = p.replace("tickFormatter","")
 ######
 header = re.findall('[0-9][0-9][0-9][0-9]', p, re.DOTALL)
 ######
    
 ######
 td = soup.find_all("table")[0].find_all("td")
 cnt = 0
 pe_hist = []
 for i in td:
  try:
    pe_hist.append(float(i.text))
    cnt+=1
    if(cnt==10):
      break
  except:
    if(i.text=="—"):
      cnt+=1
      pe_hist.append(np.NaN)
 ######
 return pd.DataFrame({"P/E":pe_hist},index=header)


#Ke.py
import requests
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np


import json
import datetime

from enum import Enum

from datetime import datetime as dt, timedelta
from dateutil.relativedelta import relativedelta

class BondType(Enum):
 USTreasury10Year = "USTreasury10Year"
 ThaiGovernmentBond10Year   = "ThaiGovernmentBond10Year"
 USTreasury10YearAnd2Year = "USTreasury10YearAnd2Year"



def validate(date_string):
 date_format = '%Y-%m-%d'
 try:
  date_obj = datetime.datetime.strptime(date_string, date_format)
  return 1
 except ValueError:
  print("Incorrect data format, should be YYYY-MM-DD")
  return 0


def thaibma_series(date="",bondtype=BondType.ThaiGovernmentBond10Year,viewlog=True):
 #date = "2020-01-01"
 
 d = date.split("-")   
 date = dt(int(d[0]),int(d[1]),int(d[2]))#startdate
 stop = dt.now()
 
 

 logYeild = []
 dateIndex = []
 for i in range(365*10): #เดาว่าไม่เกิน 1000 เดือน
  if((stop.month==date.month) & (stop.year==date.year) & (stop.day==date.day) ):
    break
    
  try:      
   dt_str = date.strftime('%Y-%m-%d')

   k = thaibma(dt_str,bondtype=bondtype)
   y = float(k)
 
   logYeild.append(float(k))
   dateIndex.append(date)

   date = date + relativedelta(days=+1)
    
   print(".",end="")
    
  except:
    dt_str = date.strftime('%Y-%m-%d')
    print("error"+dt_str,end=" ")
    date = date + relativedelta(days=+1)
    
 print("End.",end="")   
 df = pd.DataFrame({bondtype.value:logYeild},index=dateIndex)
 return df








#http://www.thaibma.or.th/EN/Market/YieldCurve/Government.aspx
def thaibma(date="",bondtype="",viewlog=True):  
 if(date==""):
    date = datetime.datetime.now().strftime("%Y-%m-%d")

 if(validate(date)==0):
     return 0
 
 if(bondtype==""):
    bondtype = BondType.USTreasury10Year
 
 if(not isinstance(BondType.USTreasury10Year,BondType)):
    print("dataType Not Match")
    return 0,0
  
 bondtype_ = bondtype.value
  
 if(bondtype_=="USTreasury10Year" or bondtype_=="USTreasury10YearAnd2Year"): 
    jsonKey = "USTreasury"  
    url = "http://www.thaibma.or.th/yieldcurve/uszyc/"+date

 if(bondtype_=="ThaiGovernmentBond10Year"):
    jsonKey = "Curve"  
    url = "http://www.thaibma.or.th/yieldcurve/gov/"+date

 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
 r = requests.get(url,headers=headers)
 r.content
 soup = BeautifulSoup(r.content, "lxml") 
    
 data  = json.loads(soup.text)  
 try: 
  if(bondtype_=="USTreasury10Year"):
   for i in data[jsonKey]:     
    if((str(i['X']))=="10.0"):
     return ((float(i['Y']))/100)
  
  if(bondtype_=="ThaiGovernmentBond10Year"):
   for i in data[jsonKey]:     
    if((str(i['X']))=="10.0"):
     return ((float(i['Y']))/100)

  if(bondtype_=="USTreasury10YearAnd2Year"):  
    p = 0
    k2 = 0 
    k10 = 0
    for i in data[jsonKey]:  
      if((str(i['X']))=="2.0"):  
         p = p+1  
         k2 = ((float(i['Y']))/100)
      if((str(i['X']))=="10.0"):
         p = p+1 
         k10 = ((float(i['Y']))/100)
      if(p==2):
         return {"USTreasury2Year":round(k2,4),"USTreasury10Year":round(k10,4)}
 except:
    if(viewlog==True):
      print("ไม่พบข้อมูล",end=" ")
    return 0


#http://www.thaibma.or.th/EN/Market/YieldCurve/Government.aspx
def governmentBond10year(date=""):  
 if(date==""):
   date = datetime.datetime.now().strftime("%Y-%m-%d")

 if(validate(date)==0):
     return 0
    
 url = "http://www.thaibma.or.th/yieldcurve/gov/"+date
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
 r = requests.get(url,headers=headers)
 r.content
 soup = BeautifulSoup(r.content, "lxml") 
    
 data  = json.loads(soup.text)   
 for i in data['Curve']:     
  if((str(i['X']))=="10.0"):
    return ((float(i['Y']))/100)
 
 print("ไม่พบข้อมูล")
 return 0


def findIndexData(strfind,g_data):
  indexat = 0
  cnt = 0
  for i in (g_data):
    if(strfind in i.text):
        indexat = cnt
    cnt+=1
    
  return indexat  

def beta(symbol):
 #ป้องกันพวก &
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
 mrkCap = np.NAN
 symbolK = symbol.replace('&', '%26')
 r = requests.get("https://www.set.or.th/set/factsheet.do?symbol="+symbolK+"&ssoPageId=3&language=th&country=TH",headers=headers)
 r.content
 soup = BeautifulSoup(r.content, "lxml")  

 g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})#[2].find_all("td")
 if(len(g_data)==0):
   print("ไม่พบข้อมูล")  
   return 0
 
 ##########################################
 indexat = findIndexData("ข้อมูลสถิติ",g_data)
 tr_ds = g_data[indexat].find_all("tr")
 ##########################################
    
    
 th_ds = []   
 data_ds = []
 for i in tr_ds:
    tddata = i.find_all("td")
    #print(tddata[0].text)
    if("ข้อมูลสถิติ" in tddata[0].text):
       for k in tddata:
          th_ds.append(k.text)
    if("Beta" in tddata[0].text):
       z = 0 
       for k in tddata:
          m = k.text  
          if(z>0):  
            if("-" in m):
                print("ไม่พบข้อมูล")
                return 0
            m = float(m.replace(",",""))
          data_ds.append(m)     
          z+=1
            
 ##########################################
 df = pd.DataFrame({"ind":th_ds,"ds":data_ds}).T
 df.columns = th_ds   
 df = df.drop(df[df["ข้อมูลสถิติ"]=="ข้อมูลสถิติ"].index)
 df = df.set_index("ข้อมูลสถิติ")
 return df  


import pandas as pd
import numpy as np

def semideviation(r):
    """
    Returns the semideviation aka negative semideviation of r
    r must be a Series or a DataFrame
    """
    is_negative = r < 0
    return r[is_negative].std(ddof=0)

def var_historic(r, level=5):
    """
    Returns the historic Value at Risk at a specified level
    i.e. returns the number such that "level" percent of the returns
    fall below that number, and the (100-level) percent are above
    """
    if isinstance(r, pd.DataFrame):
        return r.aggregate(var_historic, level=level)
    
    elif isinstance(r, pd.Series):
        return -np.percentile(r, level)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")


def cvar_historic(r, level=5):
    """
    Computes the Conditional VaR of Series or DataFrame
    """
    if isinstance(r, pd.Series):
        is_beyond = r <= -var_historic(r, level=level)
        return -r[is_beyond].mean()
    elif isinstance(r, pd.DataFrame):
        return r.aggregate(cvar_historic, level=level)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")

def drawdown_process(return_series: pd.Series):
    """Takes a time series of asset returns.
       returns a DataFrame with columns for
       the wealth index, 
       the previous peaks, and 
       the percentage drawdown
    """
    wealth_index = 1000*(1+return_series).cumprod()
    previous_peaks = wealth_index.cummax()
    drawdowns = (wealth_index - previous_peaks)/previous_peaks
    return pd.DataFrame({"Drawdown": drawdowns})

def drawdown(r):
    if isinstance(r, pd.DataFrame):
        return r.aggregate(lambda r: drawdown(r).Drawdown.min())
    elif isinstance(r, pd.Series):
        return drawdown_process(r)
    else:
        raise TypeError("Expected r to be a Series or DataFrame")   


def annualize_rets(r, periods_per_year):
    """
    Annualizes a set of returns
    We should infer the periods per year
    but that is currently left as an exercise
    to the reader :-)
    """
    compounded_growth = (1+r).prod()
    n_periods = r.shape[0]
    return compounded_growth**(periods_per_year/n_periods)-1


def annualize_vol(r, periods_per_year):
    """
    Annualizes the vol of a set of returns
    We should infer the periods per year
    but that is currently left as an exercise
    to the reader :-)
    """
    return r.std()*(periods_per_year**0.5)


def sharpe_ratio(r, riskfree_rate, periods_per_year):
    """
    Computes the annualized sharpe ratio of a set of returns
    """
    # convert the annual riskfree rate to per period
    rf_per_period = (1+riskfree_rate)**(1/periods_per_year)-1
    excess_ret = r - rf_per_period
    ann_ex_ret = annualize_rets(excess_ret, periods_per_year)
    ann_vol = annualize_vol(r, periods_per_year)
    return ann_ex_ret/ann_vol


def gbm(n_years = 10, n_scenarios=1000, mu=0.07, sigma=0.15, steps_per_year=12, s_0=100.0):
    """
    Evolution of Geometric Brownian Motion trajectories, such as for Stock Prices
    :param n_years:  The number of years to generate data for
    :param n_paths: The number of scenarios/trajectories
    :param mu: Annualized Drift, e.g. Market Return
    :param sigma: Annualized Volatility
    :param steps_per_year: granularity of the simulation
    :param s_0: initial value
    :return: a numpy array of n_paths columns and n_years*steps_per_year rows
    """
    # Derive per-step Model Parameters from User Specifications
    dt = 1/steps_per_year
    n_steps = int(n_years*steps_per_year) + 1
    rets_plus_1 = np.random.normal(loc=(mu*dt+1), scale=(sigma*np.sqrt(dt)), size=(n_steps, n_scenarios))
    rets_plus_1[0] = 1
    prices = s_0*pd.DataFrame(rets_plus_1).cumprod()
    return prices

def inst_to_ann(r):
    """
    Convert an instantaneous interest rate to an annual interest rate
    """
    return np.expm1(r)

def ann_to_inst(r):
    """
    Convert an instantaneous interest rate to an annual interest rate
    """
    return np.log1p(r)

def cir(n_years = 10, n_scenarios=1, a=0.05, b=0.03, sigma=0.05, steps_per_year=12, r_0=None):
    """
    Generate random interest rate evolution over time using the CIR model
    b and r_0 are assumed to be the annualized rates, not the short rate
    and the returned values are the annualized rates as well
    """
    if r_0 is None: r_0 = b 
    r_0 = ann_to_inst(r_0)
    dt = 1/steps_per_year
    num_steps = int(n_years*steps_per_year) + 1 # because n_years might be a float
    
    shock = np.random.normal(0, scale=np.sqrt(dt), size=(num_steps, n_scenarios))
    rates = np.empty_like(shock)
    rates[0] = r_0
    for step in range(1, num_steps):
        r_t = rates[step-1]
        d_r_t = a*(b-r_t)*dt + sigma*np.sqrt(r_t)*shock[step]
        rates[step] = abs(r_t + d_r_t) # just in case of roundoff errors going negative
        
    return pd.DataFrame(data=inst_to_ann(rates), index=range(num_steps))


import math
def discount(t, r):
    """
    Compute the price of a pure discount bond that pays a dollar at time period t
    and r is the per-period interest rate
    returns a |t| x |r| Series or DataFrame
    r can be a float, Series or DataFrame
    returns a DataFrame indexed by t
    """
    discounts = pd.DataFrame([(r+1)**-i for i in t])
    discounts.index = t
    return discounts

def pv(flows, r):
    """
    Compute the present value of a sequence of cash flows given by the time (as an index) and amounts
    r can be a scalar, or a Series or DataFrame with the number of rows matching the num of rows in flows
    """
    dates = flows.index
    discounts = discount(dates, r)
    return discounts.multiply(flows, axis='rows').sum()

def bond_cash_flows(maturity, principal=100, coupon_rate=0.03, coupons_per_year=12):
    """
    Returns the series of cash flows generated by a bond,
    indexed by the payment/coupon number
    """
    n_coupons = round(maturity*coupons_per_year)
    coupon_amt = principal*coupon_rate/coupons_per_year
    coupons = np.repeat(coupon_amt, n_coupons)
    coupon_times = np.arange(1, n_coupons+1)
    cash_flows = pd.Series(data=coupon_amt, index=coupon_times)
    cash_flows.iloc[-1] += principal
    return cash_flows
    
def bond_price(maturity, principal=100, coupon_rate=0.03, coupons_per_year=12, discount_rate=0.03):
    """
    Computes the price of a bond that pays regular coupons until maturity
    at which time the principal and the final coupon is returned
    This is not designed to be efficient, rather,
    it is to illustrate the underlying principle behind bond pricing!
    """
    cash_flows = bond_cash_flows(maturity, principal, coupon_rate, coupons_per_year)
    return pv(cash_flows, discount_rate/coupons_per_year)





#DividendDiscountModel.py

import requests
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np

def findIndexData(strfind,g_data):
  indexat = 0
  cnt = 0
  for i in (g_data):
    if(strfind in i.text):
        indexat = cnt
    cnt+=1
    
  return indexat   

def DividendDiscountModel(symbol,year,k):
 """
    ประเมินมูลค่าหุ้นด้วย Dividend Discount Model
 
    Returns ddm,g,payout,div,eps,roe

    ddm มูลค่าหุ้น , g อัตราการเติบโต , payout ,div เงินปันผลในปีนั้นๆ , eps และ roe
    
    arg1 (str): ตัวย่อของหุ้น เช่น "aot","cpall","ptt"

    arg2 (int): year ปีเป็นพ.ศ.เช่น 2561,2560 

    arg3 (float): k อัตราผลตอบแทนที่คาดหวัง
    
 """  
 if(type(year)==int):
   year = str(year)
 
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
 symbolK = symbol.replace('&', '%26')
 r = requests.get("https://www.set.or.th/set/factsheet.do?symbol="+symbolK+"&ssoPageId=3&language=th&country=TH",headers=headers)
 r.content
 soup = BeautifulSoup(r.content, "lxml") 
    
 g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})   
    
 #### หาปันผลรวมในปีของผลประกอบการนั้นๆ ก่อน #########################
 indexat = findIndexData("การจ่ายปันผล",g_data)
 tr_ds = g_data[indexat].find_all("tr") 


 dd_ds = []
 for i in tr_ds:
  td1 = i.find_all("td")    
   
  contentStr = td1[0].text
  if("ต.ค." in td1[0].text[0:12]):
    if("ก.ย." in (td1[0].text[12:len(td1[0].text)])):
        contentStr = td1[0].text[12:len(td1[0].text)]
  
  if(year in contentStr):  
     dd_tmp = (td1[1].text)
    
     if(":" in dd_tmp):
        print("พบว่าอาจจะมีปันผลเป็นหุ้น กรุณาตรวจสอบด้วยตนเอง")
        return 0,0,0,0,0,0     
     dd_tmp = float(dd_tmp)
     dd_ds.append(dd_tmp)  
      
    
 dd_ds = np.array(dd_ds).sum()   
 if(dd_ds==0):
    print("ไม่พบข้อมูล(Dividend)")
    return 0,0,0,0,0,0
 #### จบส่วนหาปันผล #########################
    
 #### หา EPS ######################### 
 indexat = findIndexData("งบกำไรขาดทุนเบ็ดเสร็จ",g_data)
 tr_ds = g_data[indexat].find_all("tr") 

    
 #หาปีที่ต้องการว่าอยู่หลักไหน
 td1 = tr_ds[0].find_all("td")
 cnt = 0
 colyear = 0
 for i in td1:
   if(year in i.text[0:4]):
     colyear = cnt
   cnt+=1 

 if(colyear==0):
    print("ไม่พบข้อมูล(eps)")
    return 0,0,0,0,0,0
 
 eps = 0 
 for i in tr_ds:
   td1 = i.find_all("td")
   if("กำไรต่อหุ้น (บาท)" in td1[0].text):
      eps = (td1[colyear].text)

      if("N/A" in eps):
       eps = np.NaN
      else:  
       eps = float(eps)

 
 if(eps==0 or np.isnan(eps)):
    print("ไม่พบข้อมูล(eps)")
    return 0,0,0,0,0,0     
 ##### จบส่วนหา EPS ################################

 ###### หา payout #####################
 payout = (dd_ds / eps)
 ##### จบส่วนหา payout ################################ 
    
    
 ###### หา ROE #####################   
 indexat = findIndexData("อัตราส่วนทางการเงิน",g_data)
 tr_ds = g_data[indexat].find_all("tr") 
 

 #หาปีที่ต้องการว่าอยู่หลักไหน
 td1 = tr_ds[1].find_all("td")
 cnt = 0
 colyear = 0
 for i in td1:
   if(year in i.text[0:4]):
     colyear = cnt
   cnt+=1 
 
 if(colyear==0):
    print("ไม่พบข้อมูล(ROE)")
    return 0,0,0,0,0,0
 
 roe = 0
 for i in tr_ds:
   td1 = i.find_all("td")
   if("อัตราผลตอบแทนผู้ถือหุ้น (%)" in td1[0].text):
    roe = (td1[colyear].text)
    if("N/A" in roe):
       roe = np.NaN
    else:
       roe = float(roe)
       roe = roe/100 

 if(roe==0 or np.isnan(roe)):
    print("ไม่พบข้อมูล(ROE)")
    return 0,0,0,0,0,0     
 ###### จบส่วนหา ROE #####################    
 
 ###### หา g ##################### 
 g = roe*(1-payout) 
 ###### จบส่วนที่หา g #####################    
    
 ######## ประเมินมูลค่าหุ้นด้วย DDM ###   
 ddm = (dd_ds*(1+g))/(k-g)   
    
 return ddm,g,payout,dd_ds,eps,roe     


#LongTermDebt

import pandas as pd
import os
from selenium import webdriver

from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support.ui import WebDriverWait

def getLongTermDebt(symbol):
  url = "http://financials.morningstar.com/balance-sheet/bs.html?t="+symbol+"&region=tha&culture=en-US"
  #os.environ['MOZ_HEADLESS'] = '1'  
 
  options = Options()
  options.add_argument('-headless')
  
  driver = webdriver.Safari()
  #driver = webdriver.Safari(options)
  #driver = webdriver.Firefox()  #อาจต้องลง 
  # Grab the web page
  #driver.manage().timeouts().implicitlyWait(10, TimeUnit.SECONDS);
  print("Processing..",end="")  
  driver.get(url)  
  wait = WebDriverWait(driver, 5)
  print(".",end="")  
  ################
  data = []  
  header = []
  ################  
  k = driver.find_element_by_id("data_i50")
  for i in k.find_elements_by_tag_name("div"):
    data.append(float(i.text.replace(",","")))
    print(".",end="")   
    
  k = driver.find_element_by_id("Year")
  for i in k.find_elements_by_tag_name("div"):
    header.append(i.text) 
    print(".",end="")   
    
  df = pd.DataFrame({"Long Term Debt":data},index=header)
  print(".",end="") 
  driver.close()
  print(".End.",end="")  
  return df


#listSecurities.py
import requests
from bs4 import BeautifulSoup
import pandas as pd

#https://www.set.or.th/set/commonslookup.do?language=th&country=TH&prefix=NUMBER
#https://www.set.or.th/set/commonslookup.do?language=th&country=TH&prefix=A 
def listSecurities(progress=True,industry=False):
    #สร้าง index ที่จะไปโหลดข้อมูลของ set
    ABC = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    listA = []
    for i in ABC:
      listA.append(i)
    
    listA.append("NUMBER")
    
    
    cnt = 0
    for i in listA:
     headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}
     url = "https://www.set.or.th/set/commonslookup.do?language=th&country=TH&prefix="+i
     r = requests.get(url,headers)
     soup = BeautifulSoup(r.content, "lxml")

     listshare = soup.find_all("table",{"class":"table table-profile table-hover table-set-border-yellow"})
     listshare = listshare[0].find_all("tr")
    
     symbol = []
     namestock = []
     typemarket = []
     industry_ds = []
     for j in (range(1,len(listshare))): 
          k1 = listshare[j].find_all("td")[0].text.strip() 
          symbol.append(k1)
          k2 = listshare[j].find_all("td")[1].text.strip() 
          namestock.append(k2)
          k3 = listshare[j].find_all("td")[2].text.strip() 
          typemarket.append(k3)
          
          if(industry):
            k1 = k1.replace("&","%26")
            headers_detail = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}
            url_detail = "https://www.set.or.th/set/factsheet.do?symbol="+k1+"&ssoPageId=3&language=th&country=TH"
            r_detail = requests.get(url_detail,headers_detail)
            soup_detail = BeautifulSoup(r_detail.content, "lxml")
            
            ctx = soup_detail.find_all("table",{"class":"table-factsheet-padding0"})
            k4 = ctx[0].find_all("tr")[0].find_all("td")[0].text.strip()
            industry_ds.append(k4)

     if(cnt==0):  
      cnt = cnt+1 
      if(progress==True):
        print("Processing.",end="") 

      if(industry):
       sData = pd.DataFrame({"symbol":symbol,"name":namestock,"market":typemarket,"industry":industry_ds})
      else:    
       sData = pd.DataFrame({"symbol":symbol,"name":namestock,"market":typemarket})

     else:
      if(industry): 
        sTmp = pd.DataFrame({"symbol":symbol,"name":namestock,"market":typemarket,"industry":industry_ds})
      else:
        sTmp = pd.DataFrame({"symbol":symbol,"name":namestock,"market":typemarket})

      sData = sData.append(sTmp) 

     if(progress==True):   
       print(".",end="") 

    if(progress==True):
      print("Complete")
    sData = sData.set_index("symbol")
    return sData


import re
import pandas as pd
import ssl
from urllib import request, parse
from bs4 import BeautifulSoup

import os
from selenium import webdriver
from selenium.webdriver.support.ui import Select
import datetime

from IPython.display import HTML

from urllib.parse import quote 


import requests
 

pd.set_option('display.max_colwidth',-1)

def filterThai(str_data):
    return re.sub(r'[^a-zA-Z]', "", str_data)

def getValueCompanyR59(wordSearch):
 '''
 wordSearch : str เช่น เจ มาร์ท , ท่าอากาศ เป็นต้น
              ใช้คำบางส่วนของชื่อบริษัท
 '''  
 context = ssl._create_unverified_context()
 url = 'https://market.sec.or.th/public/idisc/th/r59'
 
 #data = parse.urlencode(values).encode()
 req =  request.Request(url) # this will make the method "POST"
 resp = request.urlopen(req,context=context)
 
 soup = BeautifulSoup(resp.read() , "lxml")
 
 ct = soup.find_all("select")

 for i in ct[0].find_all('option'):
   if(wordSearch in i.text):
      print(i.text)
      print(i["value"])

def make_clickable(link):
    # target _blank to open new window
    # extract clickable text to display for your link
    text = "รายละเอียด"
    #link = link
    return f'<a target="_blank" href="{link}">{text}</a>'

def getReportR59(dateFrom,dateTo,Company="",report=True,Viewmore=False):
  '''
  ใช้วันที่เป็น พ.ศ. #DD/MM/YYYY 
  ตัวอย่าง getReportR59(dateFrom="01/01/2563",dateTo="30/05/2563",Company="0000003725") 
  Company : (string) รหัสของบริษัท จะคืนค่านี้มาจาก getValueCompanyR59 
  dateFrom : (string) ฟังก์ชันนี้กำหนดค่าเป็น พศ. DD/MM/YYYY
  dateTo : (string) ฟังก์ชันนี้กำหนดค่าเป็น พศ. DD/MM/YYYY
  Viewmore : (Boolean) หากกำหนดเป็น True จะคืนค่ามากกว่า 100 รายการได้

  หากใช้งานไม่ได้ ทดลองติดตั้งเพิ่มเติมด้วยคำสั่ง
  conda install -c conda-forge geckodriver
  ''' 
  if(Viewmore==True):
   d1 = dateFrom.split('/')
   dateFrom = str(int(d1[2])-543)+d1[1]+d1[0]
   
   d2 = dateTo.split('/')
   dateTo = str(int(d2[2])-543)+d2[1]+d2[0]

   url = 'https://market.sec.or.th/public/idisc/th/Viewmore/r59-2?DateType=1&DateFrom='+dateFrom+'&DateTo='+dateTo
   headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
   r = requests.post(url,headers=headers, timeout=5)
   soup = BeautifulSoup(r.content, "lxml") 
   rows = soup.find_all('table',{'table table-striped table-hover'})     
   rows = rows[0]
    
  else:
   #dateFrom = "01/01/2563"
   #dateTo = "10/01/2563"
   #Company = "0000008180"  
   #dateFrom="01/01/2563"
   #dateTo="30/04/2563"
   #Company="0000028632"
   if 'google.colab' in str(get_ipython()):
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument('--headless')
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--disable-dev-shm-usage')
    driver = webdriver.Chrome('chromedriver',chrome_options=chrome_options)
   else:   
    os.environ['MOZ_HEADLESS'] = '1'  
    driver = webdriver.Firefox()  #อาจต้องลง 
   # Grab the web page
    
   driver.get("https://market.sec.or.th/public/idisc/th/r59")
   #กำหนดค่าใน textbox
   k = driver.find_element_by_id("BSDateFrom")
   k.clear()
   k.send_keys(dateFrom)
   driver.find_element_by_id("BSDateFrom").get_attribute("value")
    
   ### Dateto
   k = driver.find_element_by_id("BSDateTo")
   k.clear()
   k.send_keys(dateTo)
   driver.find_element_by_id("BSDateTo").get_attribute("value")

   if(Company!=""):
   ### Select
    k = Select(driver.find_element_by_id("ctl00_CPH_ddlCompany"))
    k.select_by_value(Company)
    #driver.find_element_by_id("BSDateTo").get_attribute("value")
   elif(Company==""):
    k = Select(driver.find_element_by_id("ctl00_CPH_ddlCompany"))  
    k.select_by_value("")  

   ### หาปุ่ม submit และกด click
   btn_input = driver.find_element_by_id("ctl00_CPH_btSearch")
   # Then we'll fake typing into it
   btn_input.click()  

   ### เริ่ม parser
   # We can feed that into Beautiful Soup
   doc = BeautifulSoup(driver.page_source,"html.parser")  
   rows = doc.find('table', id='gPP09T01')  
  
  ############################################# 
  if("ไม่พบข้อมูล" in str(rows)):
    return 0
    #print(1)

  tr = []
  td = []
     
  for i in rows.find_all("tr"):
    td = []  
    cnt_td = 0
    k = i.find_all("td")
    for j in k:
     if(cnt_td==4): #หา date
        tmp = j.text.split("/")
        date_time_str = str(int(tmp[2])-543)+"-"+tmp[1]+"-"+tmp[0]
        dt = datetime.datetime.strptime(date_time_str, '%Y-%m-%d')
        #print(date_time_str)
        td.append(dt) 
     
     elif(cnt_td==8):
        td.append(j.a['href'])
     else:   
        td.append(j.text)   
        
     cnt_td+=1
    tr.append(td.copy())
    
  #สร้าง DataFrame  
  df = pd.DataFrame(tr)
  #จัดทรง DataFrame  
  df = df.dropna()
  
   
  df.columns = ["Company","Management","Relationship","TypesAsset","Date","Amount","AvgPrice","Methods","Remark"] 
  df["Company"] = [filterThai(j) for j in df["Company"]]
  #del df['Remark']
  
  try:
   df["Amount"] = df["Amount"].str.replace(",","").astype(float) 
  except:
   df["Amount"] = df["Amount"].str.replace(",","")

  try: 
   df["AvgPrice"] = df["AvgPrice"].str.replace(",","").astype(float)
  except:
   df["AvgPrice"] = df["AvgPrice"].str.replace(",","")

  
  if(Viewmore==False):
   driver.close()  

  #เพิ่ม link HTML รายละเอียด

  if(report==True):
   df['Remark'] = df['Remark'].apply(make_clickable)
   df = HTML(df.to_html(escape=False))
  return df  







#### แบบรายงาน R246
def getReportR246(symbolOrName,datefrom,dateto,person=False):
    '''
    pl.getReportR246(symbolOrName='',datefrom='2020-10-01',dateto='2020-10-30')
    symbolOrName : (string) ชื่อหุ้น หากไม่ระบุจะคืนค่าทุกตัวในช่วงวันนั้นๆ
    datefrom : (string) ระบุเป็นคศ. YY-MM-DD
    dateto :  (string) ระบุเป็นคศ. YY-MM-DD
    '''
    ###########
    datefrom = datefrom.replace("-","")
    dateto = dateto.replace("-","")
    
    if(person==False):
     url =  "https://market.sec.or.th/public/idisc/th/Viewmore/r246-2" \
      + "?SearchSymbol="+symbolOrName+"&ListedType=listed&DateType=1&DateFrom="+datefrom \
      +"&DateTo="+dateto
    if(person):
       person = quote(symbolOrName) #.encode('utf-8')
       url =  "https://market.sec.or.th/public/idisc/th/Viewmore/r246-2" \
          + "?SearchPerson="+person+"&ListedType=listed&DateType=1&DateFrom="+datefrom \
          +"&DateTo="+dateto  
       #print(url)
    
    ###########
    context = ssl._create_unverified_context()
    req =  request.Request(url) # this will make the method "POST"
    resp = request.urlopen(req,context=context)
    soup = BeautifulSoup(resp.read() , "lxml")
    
    ###########
    rows = soup.find('table', id='gPP10T01')
    
    ct = 0
    data = rows.find_all("tr")
    symbolDS = []
    nameDS = []
    type_rDS = []
    type_assetDS = []
    pct_beforeDS = []
    ptc_actionDS = []
    ptc_afterDS = []
    date_actionDS = []
    
    for i in range(len(data)):
     if(i>0):   
      tdData = (data[i].find_all("td"))
    
      if("ไม่พบข้อมูล" in tdData[0]):
          print(tdData[0].text)  
          return 0
            
      symbol = (tdData[0].text)
      symbolDS.append(symbol) 
    
      name = (tdData[1].text)   
      nameDS.append(name) 
    
      type_r = (tdData[2].text)
      type_rDS.append(type_r)
    
      type_asset = (tdData[3].text) 
      type_assetDS.append(type_asset)
    
      pct_before = (tdData[4].text)
      pct_beforeDS.append(pct_before)
    
      ptc_action = (tdData[5].text)
      ptc_actionDS.append(ptc_action)
    
      ptc_after = (tdData[6].text)
      ptc_afterDS.append(ptc_after)
    
      date_action = (tdData[7].text) 
      date_actionDS.append(date_action)
    
    ###########
    df = pd.DataFrame({"symbol":symbolDS,"name":nameDS,"type":type_rDS,"asset":type_assetDS,
              "pctBefore":pct_beforeDS,"pctAction":ptc_actionDS,"pctAfter":ptc_afterDS,"date":date_actionDS})
    
    return df


import starfishX as sx

keywordList = ["นาย","นาง","พญ","พ.ญ.","น.ส.","ท่านผู้หญิง","ร.ต.","mr","mrs","miss","เด็กชาย","ด.ช.",
                   "เด็กหญิง","ด.ญ.","พ.อ.","พ.ท.","พ.ต.","ร้อยตรี","ร้อยโท","ร้อยเอก","เรือตรี","เรือโท","เรือเอก"
                   "เรืออากาศเอก","พันเอก","พันโท","แพทย์","หม่อม","ม.ล.","ม.ร.ว","พลตำรวจ","พันตำรวจ","ร.ท."
                   "ว่าที่","ร.ศ.","ดร.","คุณหญิง","รองศาสตราจารย์","นพ.","น.พ.","สมเด็จพระ",]
def filterPerson(n):    
    for i in keywordList:  
     if(i in n.lower()):
        return False
    return True

def filterPersonC(n):    
    for i in keywordList:  
     if(i in n.lower()):
        return True
    return False

def filter2float(n):
    n = n.str.replace(',', '')
    n = n.astype(float)  
    return n

def getRatioPersonalShareholders(symbol):
    logAll = sx.listShareholders(symbol)
    logAll['Share'] = filter2float(logAll['Share'])

    logIn = logAll[logAll["Name"].map(filterPerson)] 
    logPerson = logAll[logAll["Name"].map(filterPersonC)] 

    inHold = logIn["Share"].sum()
    allHold = logAll["Share"].sum()
    r = inHold/allHold
    return (r),logIn,logPerson,logAll






import requests
import json 
import pandas as pd
from datetime import datetime

try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    #import config as con 
    from starfishX import config as con 
except:
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""

def getDW(symbol):
    '''
      str : symbol 
      เช่น symbol  'INTU01C2201A'   
    '''
    urlStarter = 'https://api.set.or.th/api/stock/'+symbol+'/historical-trading' 

    resp = requests.get(urlStarter)
    txt = resp.json()

    df = pd.DataFrame(txt)
    df['date'] = pd.to_datetime(df['date']) #convert str2date
    df = df.set_index('date')

    return df


#incomestatement.py
import requests
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np

 
def getIncomeStatement(symbol):
  headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
  symbolMofi = symbol.replace("&","%26")  
  r = requests.post("https://www.set.or.th/set/factsheet.do?symbol="+symbolMofi,headers=headers)
  r.content
  soup = BeautifulSoup(r.content, "lxml") 

  if len(soup.find_all("table", {"class" : "table-factsheet-padding0"}))==0:
     print("ไม่พบข้อมูล")
     return 0

  #g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[19].find_all("td")
  #findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[19].find_all("tr")
  #if not "งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in str(g_data):
  #  g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[20].find_all("td")
  #  findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[20].find_all("tr")
  

  tmpdata = soup.find_all("table", {"class" : "table-factsheet-padding0"})
  if("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[19].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[19].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[19].find_all("tr")
  elif("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[20].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[20].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[20].find_all("tr")
  elif("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[18].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[18].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[18].find_all("tr") 
  elif("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[17].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[17].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[17].find_all("tr")   
  elif("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[16].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[16].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[16].find_all("tr") 
  elif("งบกำไรขาดทุนเบ็ดเสร็จ (ลบ.)" in tmpdata[21].text):
    g_data = soup.find_all("table", {"class" : "table-factsheet-padding0"})[21].find_all("td")
    findStep = soup.find_all("table", {"class" : "table-factsheet-padding0"})[21].find_all("tr")
    
  ds  = []
  row = []
  cnt = 1
 
  #lenstep = 6    
  #if "M" in str(g_data[1]):
  #   lenstep = 7

  #ปรับโค้ดในการหาจำนวนคอลัมม์ที่แสดง เช่นกรณี vranda หุ้นใหม่จะมีคอลัมม์มาแค่ 3 คอลัมม์
  #0.15493
  lenstep = len(findStep[0].text.strip().split("\n"))+1    
  #print(lenstep)

  for i in range(0,len(g_data)-1):
    s = str(g_data[i].text.strip())
    s = s.rstrip()
    s = s.lstrip()
    
    row.append(s)
    #print(g_data[i].text.strip())
    cnt+=1
    if(cnt==lenstep): 
      ds.append(row)
      row = []
      cnt = 1
    
  df = pd.DataFrame(ds)
  df.columns = df.iloc[0]   
  
  #edit version 1.543
  df = df.drop(0)
  trim_columnsname(df) 
  df = df.set_index("งบกำไรขาดทุนเบ็ดเสร็จ(ลบ.)")
  return df


def getIncomeStatementCompare(liststock):
 ds = []
 if type(liststock) is list:
  for i in liststock:
    t = getIncomeStatement(i)#.report()
    #t = sx.trim_columnsname(t)
    ds.append(t)
    
 BusinessType = True
 for i in range(len(ds)):
  if(BusinessType==False):
    break
  for j in range(len(ds)):
   #chk = np.array_equal(ds[i][ds[i].columns[0]],ds[j][ds[j].columns[0]])
   chk = np.array_equal(ds[i].index,ds[j].index)
   BusinessType = chk
   if(BusinessType==False):
      print("Can not compare Income Statement")
      break
    
 if(BusinessType==False):
    return 0

 p = []
 #s0 = ds[0][[ds[0].columns[0]]]
 #s0 = ds[0].index 
 #p.append(s0)
 for i in range(len(ds)):
  #s0 = ds[i][[ds[i].columns[0]]]
  #s0.columns = [liststock[i]+"."+s0.columns[0]]
  s0 = ds[i][ds[i].columns[0]]  
  p.append(s0)  

 z = pd.concat(p, axis=1)
 return z



def trim_columnsname(df):
 cc = []
 for i in range(len(df.columns)):
   tmp = df.columns[i]
   tmp = tmp.replace("\xa0", "")
   tmp = tmp.replace(" ", "")  
   cc.append(tmp)
 df.columns = cc
 return df


# Squarified Treemap Layout
# Implements algorithm from Bruls, Huizing, van Wijk, "Squarified Treemaps"
#   (but not using their pseudocode)


# INTERNAL FUNCTIONS not meant to be used by the user


def pad_rectangle(rect):
    if rect["dx"] > 2:
        rect["x"] += 1
        rect["dx"] -= 2
    if rect["dy"] > 2:
        rect["y"] += 1
        rect["dy"] -= 2


def layoutrow(sizes, x, y, dx, dy):
    # generate rects for each size in sizes
    # dx >= dy
    # they will fill up height dy, and width will be determined by their area
    # sizes should be pre-normalized wrt dx * dy (i.e., they should be same units)
    covered_area = sum(sizes)
    width = covered_area / dy
    rects = []
    for size in sizes:
        rects.append({"x": x, "y": y, "dx": width, "dy": size / width})
        y += size / width
    return rects


def layoutcol(sizes, x, y, dx, dy):
    # generate rects for each size in sizes
    # dx < dy
    # they will fill up width dx, and height will be determined by their area
    # sizes should be pre-normalized wrt dx * dy (i.e., they should be same units)
    covered_area = sum(sizes)
    height = covered_area / dx
    rects = []
    for size in sizes:
        rects.append({"x": x, "y": y, "dx": size / height, "dy": height})
        x += size / height
    return rects


def layout(sizes, x, y, dx, dy):
    return (
        layoutrow(sizes, x, y, dx, dy) if dx >= dy else layoutcol(sizes, x, y, dx, dy)
    )


def leftoverrow(sizes, x, y, dx, dy):
    # compute remaining area when dx >= dy
    covered_area = sum(sizes)
    width = covered_area / dy
    leftover_x = x + width
    leftover_y = y
    leftover_dx = dx - width
    leftover_dy = dy
    return (leftover_x, leftover_y, leftover_dx, leftover_dy)


def leftovercol(sizes, x, y, dx, dy):
    # compute remaining area when dx >= dy
    covered_area = sum(sizes)
    height = covered_area / dx
    leftover_x = x
    leftover_y = y + height
    leftover_dx = dx
    leftover_dy = dy - height
    return (leftover_x, leftover_y, leftover_dx, leftover_dy)


def leftover(sizes, x, y, dx, dy):
    return (
        leftoverrow(sizes, x, y, dx, dy)
        if dx >= dy
        else leftovercol(sizes, x, y, dx, dy)
    )


def worst_ratio(sizes, x, y, dx, dy):
    return max(
        [
            max(rect["dx"] / rect["dy"], rect["dy"] / rect["dx"])
            for rect in layout(sizes, x, y, dx, dy)
        ]
    )


# PUBLIC API


def squarify(sizes, x, y, dx, dy):
    """Compute treemap rectangles.

    Given a set of values, computes a treemap layout in the specified geometry
    using an algorithm based on Bruls, Huizing, van Wijk, "Squarified Treemaps".
    See README for example usage.

    Parameters
    ----------
    sizes : list-like of numeric values
        The set of values to compute a treemap for. `sizes` must be positive
        values sorted in descending order and they should be normalized to the
        total area (i.e., `dx * dy == sum(sizes)`)
    x, y : numeric
        The coordinates of the "origin".
    dx, dy : numeric
        The full width (`dx`) and height (`dy`) of the treemap.

    Returns
    -------
    list[dict]
        Each dict in the returned list represents a single rectangle in the
        treemap. The order corresponds to the input order.
    """
    sizes = list(map(float, sizes))

    if len(sizes) == 0:
        return []

    if len(sizes) == 1:
        return layout(sizes, x, y, dx, dy)

    # figure out where 'split' should be
    i = 1
    while i < len(sizes) and worst_ratio(sizes[:i], x, y, dx, dy) >= worst_ratio(
        sizes[: (i + 1)], x, y, dx, dy
    ):
        i += 1
    current = sizes[:i]
    remaining = sizes[i:]

    (leftover_x, leftover_y, leftover_dx, leftover_dy) = leftover(current, x, y, dx, dy)
    return layout(current, x, y, dx, dy) + squarify(
        remaining, leftover_x, leftover_y, leftover_dx, leftover_dy
    )


def padded_squarify(sizes, x, y, dx, dy):
    """Compute padded treemap rectangles.

    See `squarify` docstring for details. The only difference is that the
    returned rectangles have been "padded" to allow for a visible border.
    """
    rects = squarify(sizes, x, y, dx, dy)
    for rect in rects:
        pad_rectangle(rect)
    return rects


def normalize_sizes(sizes, dx, dy):
    """Normalize list of values.

    Normalizes a list of numeric values so that `sum(sizes) == dx * dy`.

    Parameters
    ----------
    sizes : list-like of numeric values
        Input list of numeric values to normalize.
    dx, dy : numeric
        The dimensions of the full rectangle to normalize total values to.

    Returns
    -------
    list[numeric]
        The normalized values.
    """
    total_size = sum(sizes)
    total_area = dx * dy
    sizes = map(float, sizes)
    sizes = map(lambda size: size * total_area / total_size, sizes)
    return list(sizes)


def plot(
    sizes,
    norm_x=100,
    norm_y=100,
    color=None,
    label=None,
    value=None,
    ax=None,
    pad=False,
    bar_kwargs=None,
    text_kwargs=None,
    **kwargs
):
    """Plotting with Matplotlib.

    Parameters
    ----------
    sizes
        input for squarify
    norm_x, norm_y
        x and y values for normalization
    color
        color string or list-like (see Matplotlib documentation for details)
    label
        list-like used as label text
    value
        list-like used as value text (in most cases identical with sizes argument)
    ax
        Matplotlib Axes instance
    pad
        draw rectangles with a small gap between them
    bar_kwargs : dict
        keyword arguments passed to matplotlib.Axes.bar
    text_kwargs : dict
        keyword arguments passed to matplotlib.Axes.text
    **kwargs
        Any additional kwargs are merged into `bar_kwargs`. Explicitly provided
        kwargs here will take precedence.

    Returns
    -------
    matplotlib.axes.Axes
        Matplotlib Axes
    """

    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()

    if color is None:
        import matplotlib.cm
        import random

        cmap = matplotlib.cm.get_cmap()
        color = [cmap(random.random()) for i in range(len(sizes))]

    if bar_kwargs is None:
        bar_kwargs = {}
    if text_kwargs is None:
        text_kwargs = {}
    if len(kwargs) > 0:
        bar_kwargs.update(kwargs)

    normed = normalize_sizes(sizes, norm_x, norm_y)

    if pad:
        rects = padded_squarify(normed, 0, 0, norm_x, norm_y)
    else:
        rects = squarify(normed, 0, 0, norm_x, norm_y)

    x = [rect["x"] for rect in rects]
    y = [rect["y"] for rect in rects]
    dx = [rect["dx"] for rect in rects]
    dy = [rect["dy"] for rect in rects]

    ax.bar(
        x, dy, width=dx, bottom=y, color=color, label=label, align="edge", **bar_kwargs
    )

    if not value is None:
        va = "center" if label is None else "top"

        for v, r in zip(value, rects):
            x, y, dx, dy = r["x"], r["y"], r["dx"], r["dy"]
            ax.text(x + dx / 2, y + dy / 2, v, va=va, ha="center", **text_kwargs)

    if not label is None:
        va = "center" if value is None else "bottom"
        for l, r in zip(label, rects):
            x, y, dx, dy = r["x"], r["y"], r["dx"], r["dy"]
            ax.text(x + dx / 2, y + dy / 2, l, va=va, ha="center", **text_kwargs)

    ax.set_xlim(0, norm_x)
    ax.set_ylim(0, norm_y)
   
    return ax



import requests
import pandas as pd
from bs4 import BeautifulSoup

def getHolidayMarket(year):
   '''
   year : int เป็นปี คศ.เช่น 2020 โดยเริ่มต้นจาก 1992
   ref : https://www.bot.or.th/Thai/FinancialInstitutions/FIholiday/Pages/2021.aspx
   '''
   ######## 
   url = "https://www.bot.or.th/Thai/FinancialInstitutions/FIholiday/Pages/"+str(year)+".aspx"
    
   headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
   r = requests.post(url,headers=headers, timeout=5)
   soup = BeautifulSoup(r.content, "lxml")   
   ########
   idzone = soup.find_all("div",{"class","ms-rtestate-field"})  
   if(len(idzone)>1):#support ปี 2014
         idzone[0] = idzone[-1]
   tb = (idzone[0].find_all("tr"))
    
   '''tb = soup.find_all("table",{"class":"bot-rteTable-mytable2"})

   if(len(tb)==1):
     tb = tb[0].find_all("tr") 
    
   if(len(tb)==0): #support ปี 1998
     tb = soup.find_all("table",{"class":"bot-rteTable-mytable"}) 
     tb = tb[1]
     tb = tb.find_all("tr")   
        
   if(not "มกราคม" in tb): #support ปี 2011
     idzone = soup.find_all("div",{"class","pageContent"})
     if(len(idzone)==0): #support ปี 2013
       idzone = soup.find_all("div",{"class","ms-rtestate-field"})  
       if(len(idzone)>1):#support ปี 2014
         idzone[0] = idzone[-1]
     tb = (idzone[0].find_all("tr"))   '''   
   ########
   monthTH = ["มกราคม","กุมภาพันธ์","มีนาคม","เมษายน", \
             "พฤษภาคม","มิถุนายน","กรกฎาคม","สิงหาคม","กันยายน","ตุลาคม","พฤศจิกายน","ธันวาคม"]
   monthEN = ["01","02","03","04","05","06","07","08","09","10","11","12"]
   #print(soup)
   ds = []
   holiday = []
   for tbi in tb:
     try:
       data = tbi.find_all("td") 
       mt = data[3].text.strip()
       mt = mt.replace("​","")
       d = data[2].text.strip().replace("​","") 
     except:
       continue  
     
     if(len(d)==1):
        d = "0"+d  
     #k = [i for i in range(len(monthTH)) if monthTH[i]==mt][0]
     p = 0
     for i in range(len(monthTH)):
        if(monthTH[i]==mt):
           k = i
           p = 1
        
     if(p==0):
       continue

     m = monthEN[k]
     date = str(year)+"-"+m+"-"+d 
     whoday = data[4].text.replace("\r\n","").replace("\n","")
        
     if("ยกเลิก" in whoday):
        continue
        
     ds.append(date)
     holiday.append(whoday)

   return pd.DataFrame({"ds":pd.to_datetime(ds),"holiday":holiday}) 


#marketview.py
import pandas as pd
import requests
import numpy as np
from enum import Enum
import matplotlib.pyplot as plt

###################### ส่วนที่ใช้ Library ภายใน ###################
try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    from starfishX import config as con 
    #print("Dev")
except:
    #print("Debug")
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""

# สำหรับ getMemberOfIndex    #starfishX Production #__init__ debug Mode
#from starfishX.indexMarket import indexMarket
exec("from "+prefix+"indexMarket import indexMarket")
###################### END ส่วนที่ใช้ Library ภายใน ###################

 

class Market(Enum):
 PE = "PE"
 Index = "Index"

def processHistMarket(MarketType):
 typeH = MarketType.value   
 ##############################################   
 if(typeH=="PE"):
  url = "https://www.set.or.th/static/mktstat/Table_PE.xls"
  filename = "Table_PE.html"

 if(typeH=="Index"):
  url = "https://www.set.or.th/static/mktstat/Table_Index.xls"
  filename = "Table_Index.html"
 
 # download the file contents in binary format
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
   
 r = requests.get(url,headers=headers)

 # open method to open a file on your system and write the contents
 with open(filename, "wb") as code:
    code.write(r.content)
 ##############################################

 ##############################################   
 df = pd.read_html(filename) 
 ##############################################
 if(typeH=="Index"):
  rp = pd.DataFrame({ "SET":df[1].columns[1],"SET50":df[1].columns[2]
                   ,"SET100":df[1].columns[3],"sSET":df[1].columns[4]
                   ,"SETCLMV":df[1].columns[5],"SETHD":df[1].columns[6]
                   ,"SETTHSI":df[1].columns[7],"SETWB":df[1].columns[8]
                   ,"mai":df[1].columns[9]},index=df[1].columns[0])

  rp = rp.drop(rp.index[0])
  rp = rp.drop(rp.index[0])

  for i in rp.columns:
    rp[i] = rp[i].str.replace(",","")
    rp[i] = pd.to_numeric(rp[i], errors='coerce')
    #rp[i] = rp[i].astype(float)
    
  rp = rp.sort_index()   
  rp.index = pd.to_datetime(rp.index)
    
 ##############################################   
 if(typeH=="PE"):
  rp = df[1]
  rp.columns = rp.iloc[0]
  rp = rp.set_index("Month-Year")
  rp = rp.drop(rp.index[0])
  rp.index = pd.to_datetime(rp.index)

  rp = rp.sort_index()
  for i in rp.columns:
    rp[i] = rp[i].astype(float)  
    
 ##############################################
 return rp

##############################################
##############################################
def marketview(indexType,start="1900-01-01",viewplot=True):
 """
    Returns Close Index,PE : Close Index ตามประเภทของ Index นั้นๆ , PE ของ Index
    
    arg1 (indexMarket) : ประเภทของดัชนีที่เราต้องการ เช่น sx.indexMarket.SET , sx.indexMarket.SET50 เป็นต้น
    
    arg2 (str)    : start ปี โดยเป็นปี พ.ศ. เช่น "2010-01-01" YYYYMMDD 

    arg3 (Boolean) : viewplot แสดงกราฟผลของราคาปิดดัชนีและ PE หรือไม่ 
    
 """     
 if(not isinstance(indexType, indexMarket)):
    print("Data Type indexMarket Not Match")
    return 0

 marketType = indexType.value
    
 p1 = processHistMarket(Market.Index)
 p2 = processHistMarket(Market.PE)
 
 rp = pd.concat([p1[marketType], p2[marketType]],axis=1)
 rp.columns = ["index","pe"]
    
 rp = rp.dropna()  
 
 if(start!=""):
    rp = rp[rp.index>=start]   

 if(viewplot==True):   
  plt.subplot(211)
  rp['index'].plot(title=marketType.upper()+" INDEX",figsize=(20,10))
  plt.subplot(212)
  rp['pe'].plot(title=marketType.upper()+" INDEX : P/E",figsize=(20,10))
 
 return rp[rp.index>=start] 




def marketViewForeignTrade(indexType,start="",viewplot=True,widthP=20,heightP=6):
 """
    Returns DataFrame : ปริมาณซื้อและขายของนักลงทุนต่างชาติแยกตามประเภทตลาด SET และ mai
    
    arg1 (indexMarket) : ประเภทของดัชนีที่เราต้องการ เช่น sx.indexMarket.SET , sx.indexMarket.mai โดยใช้ได้เพียง 2 ตลาดเท่าน้ัน
    
    arg2 (str)    : year ปี โดยเป็นปี พ.ศ. เช่น "2010-01-01" YYYYMMDD 

    arg3 (Boolean) : viewplot แสดงกราฟผลของราคาปิดดัชนีและ PE หรือไม่ 
    
 """  
 #ตรวจ data type   
 if(not isinstance(indexType, indexMarket)):
    print("Data Type indexMarket Not Match")
    return 0
 
 #support แค่ set และ mai   
 typeM = indexType.value
    
 if(not (typeM=="SET" or typeM=="mai")): 
  print("NOT SUPPORT., "+typeM)  
  return 0 
        
 filename = "Table_ForeignTrade.html"
 url = "https://www.set.or.th/static/mktstat/Table_ForeignTrade.xls"
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}   
   
 r = requests.get(url,headers=headers)

 # open method to open a file on your system and write the contents
 with open(filename, "wb") as code:
    code.write(r.content)
 ##############################################
 
 ##############################################   
 df = pd.read_html(filename) 
 ##############################################
 rp = df[1]
 rp = rp.drop(rp.index[0])
 rp = rp.drop(rp.index[0])  
 rp.columns = ["Date","SET.BUY","SET.SELL","mai.BUY","mai.SELL"]   
 rp["Date"] = pd.to_datetime(rp["Date"])
 rp = rp.set_index("Date")   
 rp = rp.sort_index()
   
 ##############################################
 # clean กับ ปรับ data type เอาพวกค่าว่างออก   
 for i in rp.columns:
    rp[i] = rp[i].str.replace(",","")
    rp[i] = pd.to_numeric(rp[i], errors='coerce')
 
 ##############################################
 rp['SET.NET'] = rp['SET.BUY']-rp['SET.SELL']
 rp['mai.NET'] = rp['mai.BUY']-rp['mai.SELL']
 
 ##############################################
 if(viewplot==True):
   #ป้อง date format ที่มีชั่วโมงนาที 
   rp['DateReport'] = pd.to_datetime(rp.index)
   rp['DateReport'] = rp['DateReport'].dt.strftime('%Y-%m') 
   
   typeMC = typeM+".NET" 

   if(start!=""):  
     rp[rp.index>=start][["DateReport",typeMC]]. \
        plot(figsize=(widthP,heightP),kind="bar",x="DateReport",y=[typeMC],title="Foreign Trade on "+typeM) 
   else:
     rp[["DateReport",typeMC]]. \
        plot(figsize=(widthP,heightP),kind="bar",x="DateReport",y=[typeMC],title="Foreign Trade on "+typeM)
        
 return rp       


import ssl
from urllib import request, parse
from bs4 import BeautifulSoup

import os
from selenium import webdriver

import pandas as pd
import requests

def getFreeCashFlow(symbol):
    
 #symbol = "splai"
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 symbolMofi = symbol.replace("&","%26")
 url = "https://www.set.or.th/set/companyfinance.do?type=cashflow&symbol="+symbolMofi+"&language=th&country=TH"

 r = requests.get(url,headers)
 soup = BeautifulSoup(r.content, "lxml")

 data = soup.find_all("table",{"class":"table table-hover table-info"})  
 #data = soup.find_all("table",{"class":"table table-info-wrap"})
 #### หา index #### content อยู่ช่องไหน
 i_index = 0
 for i in data:
  if("งบกระแสเงินสด" in i.text):
      m_index = i_index
      break
  i_index+=1

 #######
 
 td = data[m_index].find_all("td")

 #######
 cnt = 0
 lb = []
 value = []
 vaS = 0
 vaE = 0
 k = 0
 multirow = 0
 
 for i in td:
  text = i.text.replace(" ","")  
  ###### 
  if(("เงินสดสุทธิได้มาจาก" in text) and ("กิจกรรมดำเนินงาน" in text) ):
     k = float(td[cnt+1].text.replace(",",""))
     p1 = k 
     lb.append(text)
     value.append(k)
         
  if(("เงินสดจ่ายจากการซื้อที่ดิน" in text)): #เงินสดจ่ายจากการซื้อที่ดิน อาคาร และอุปกรณ์
     k = float(td[cnt+1].text.replace(",",""))
     p1 += k 
     lb.append(text)
     value.append(k) 
       
  if(("สินทรัพย์ไม่มีตัวตน(เพิ่มขึ้น)ลดลง" in text) and not("\xa0" in text) ):  #สินทรัพย์ไม่มีตัวตน (เพิ่มขึ้น) ลดลง
     lbS = text
     vaS = float(td[cnt+1].text.replace(",",""))
     multirow = 1   
        
  if(("สินทรัพย์ไม่มีตัวตน(เพิ่มขึ้น)" in text) and ("\xa0" in text) ):  #สินทรัพย์ไม่มีตัวตน (เพิ่มขึ้น) ลดลง
     lbM = text
     vaM = float(td[cnt+1].text.replace(",",""))
     multirow = 2  
 
   
  #######
  cnt+=1 

 
 if(multirow==1):
     p1 += vaS 
     lb.append(lbS)
     value.append(vaS) 
 elif(multirow==2):
     p1 += vaM 
     lb.append(lbM)
     value.append(vaM)
 
        
 lb.append("กระแสเงินสดอิสระ")
 value.append(p1)
 df = pd.DataFrame({"Label":lb,"Value":value})
 return df   


def getFreeCashFlowWSJ(symbol,period="ANNUAL"):
 '''
 symbol : (str) สัญลักษณ์ชื่อย่อของหุ้น
 period : (str) กรอบระยะเวลา มีให้เลือกสองอย่าง ANNUAL และ QUARTERLY
 '''   
 #symbol = "dtac"
 #period = "ANNUAL"
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 symbolMofi = symbol.replace("&","%26")
 url = "https://www.wsj.com/market-data/quotes/TH/XBKK/"+symbolMofi+"/financials"
 
 print("Processing.",end="")
 os.environ['MOZ_HEADLESS'] = '1'
 print(".",end="")
 driver = webdriver.Firefox()  #อาจต้องลง 
 # Grab the web page
 print(".",end="")   
 driver.get(url)
 print(".",end="")

 test = driver.find_elements_by_xpath("//div[@class='cr_mod_full cr_data cr_data_chart cr_financials cr_financials_cash module']")
 print(".",end="")
 for i in test:
   print(".",end="")     
   for a_data in i.find_elements_by_tag_name("li"):
     #print(a_data.text)
     if(a_data.text==period):
         a_data.click()
     if(a_data.text==period):
         a_data.click()   
        
 #################  
 print(".",end="")
 
 passLabel = 0
 value = 0
 passLabelPershare = 0
 valuePershare = 0
 cnt = 0   
 for event in driver.find_elements_by_class_name("cr_financials_table"):
  for td_data in event.find_elements_by_tag_name("td"):  
    #if("Free Cash Flow" in td_data.text):  
    #cnt+=1
    #print("cnt"+str(cnt))
    #print(".",end="")
    if(("Free Cash Flow" in td_data.text) and not("Per Share" in td_data.text)):
        passLabel = 1
        #print("a",td_data.text)
        print(".",end="")
    else:
        if(passLabel==1): #find next td
          value = td_data.text
          #print("b",td_data.text)
          print(".",end="")  
          passLabel = 2
        
    if("Free Cash Flow Per Share" in td_data.text):
        passLabelPershare = 1
        #print("c",td_data.text)
        print(".",end="")
    else: 
        if(passLabelPershare==1):
          valuePershare = td_data.text
          #print("d",td_data.text)
          print(".",end="")  
          passLabelPershare = 2
          break
 #print(3) 
 print(".End.",end="")
 driver.close() 
 return pd.DataFrame({"Free Cash Flow":[value],"Free Cash Flow/Share":[valuePershare]},index=[symbol.upper()])


#openInterest.py
import requests
from bs4 import BeautifulSoup
import pandas as pd

from datetime import datetime
def str2Date(dt):
 k = dt.split(" ")
 if(k[1]=="ธ.ค."):
    k[1]=12
 elif(k[1]=="พ.ย."):
    k[1]=11    
 elif(k[1]=="ต.ค."):
    k[1]=10
 elif(k[1]=="ก.ย."):
    k[1]=9
 elif(k[1]=="ส.ค."):
    k[1]=8
 elif(k[1]=="ก.ค."):
    k[1]=7
 elif(k[1]=="มิ.ย."):
    k[1]=6
 elif(k[1]=="พ.ค."):
    k[1]=5
 elif(k[1]=="เม.ย."):
    k[1]=4
 elif(k[1]=="มี.ค."):
    k[1]=3
 elif(k[1]=="ก.พ."):
    k[1]=2    
 elif(k[1]=="ม.ค."):
    k[1]=1          
 k[2] = int(k[2])-543
 k = str(k[0])+"/"+str(k[1])+"/"+str(k[2])
 dt = datetime.strptime(k,'%d/%m/%Y')
 return dt

def listSSF(where=""): #,yesterday=False
 """
 ดูข้อมูลของ SSF ในซีรีย์ที่มีสถานะอยู่
 where : (string) คัดกองเฉพาะหุ้นแม่ที่สนใจ
 yesterday : (Boolean) ขอข้อมูลวันก่อนหน้าแทน
 """  
 #url = "https://www.tfex.co.th/tfex/dailyMarketReport.html?locale=en_US&marketListId=SF&instrumentId=&selectedDate=C&periodView=A"
 url = "https://www.tfex.co.th/tfex/dailyMarketReport.html?locale=en_US&marketListId=SF&instrumentId=&selectedDate=C&periodView=A"
 #if(yesterday==True):
 #    url = 'https://www.tfex.co.th/tfex/dailyMarketReport.html?locale=en_US'

 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}
  #https://www.tfex.co.th/th/products/stock-mktdata.html
 r = requests.get(url,headers)
 soup = BeautifulSoup(r.content, "lxml")    
 #print(soup)
 cnt = 0
 underlying_ds = []
 ssf_ds = []
 underlying_name = ""
 if(len(soup.find_all("tbody"))==0):
    print("no content")    
    return 0
 for rowi,row in enumerate(soup.find_all("tbody")[0].find_all("tr"),0):
    if(rowi>0):
     c = row.find_all("td")
     #for j,k in enumerate(c,0):
     tmp = c[0].text.strip()   
     #--- fixbug 18 มิถุนายน 2020 --- 
     #--- เพราะรู้สึกมีการแก้หน้า HTML ของเพจ ทำให้ listSSF ออกมาผิดพลาด ---
     if( not(("Total" in tmp) or ("Futures" in tmp)) ):
        #if(cnt%4==0):
           underlying_name = tmp.replace("Futures","").strip()
        #if(cnt%4!=0):
           ssf_ds.append(tmp)
           underlying_ds.append(underlying_name)
        #cnt+=1
 df = pd.DataFrame({"underlying":underlying_ds},index=ssf_ds)
 if(where!=""):   
   return df[df["underlying"].str.upper().str.contains(where.upper())].index.tolist() 
 return df

def OpenInterestContractsProcess(symbol):
  url = "https://www.tfex.co.th/tfex/historicalTrading.html?locale=th_TH&symbol="+symbol+"&decorator=excel&locale=th_TH"
  headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}
  
  r = requests.get(url,headers)
  soup = BeautifulSoup(r.content, "lxml")
    
  ct = soup.find_all("table")[2]
  rowData = ct.find_all("tr")
 
  date_ds = []
  data_oi = []
  for td in rowData:
    d = td.find_all("td")
    if((d[0].text.strip()!="วันที่") and (d[0].text.strip()!="รวมทั้งหมด")):
        date_ds.append(str2Date(d[0].text.strip()))
        oi = d[9].text.strip().replace(",","").replace("-","0")
        data_oi.append(float(oi))
        
  df = pd.DataFrame({"OI":data_oi},index=date_ds)
  return df

def OpenInterestContracts(listSymbol):
   """
   listSymbol : (list) ส่งค่าซีรีย์ SSF ที่ต้องการ ตัวอย่าง การดึงค่าจาก sx.listSSF(where='KCE')
                df = sx.OpenInterestContracts(["KCEU20","KCEZ20","KCEH21","KCEM21"])
   """
   if(type(listSymbol)==str):
      return OpenInterestContractsProcess(listSymbol)
   if(type(listSymbol)==list): 
      datasource = []
      for i in listSymbol:
         datasource.append(OpenInterestContractsProcess(i))
            
      k = pd.concat(datasource, axis=1, join='outer')#.sum(axis=1) 
      m = pd.DataFrame({"OI":k.sum(axis=1)})
      return m   


######### moc.py
try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    from starfishX import config as con 
    #print("Dev")
except:
    #print("Debug")
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""

    

import urllib
import ssl
from bs4 import BeautifulSoup

import pandas as pd
import calendar
import time
from urllib import request, parse

import datetime
from dateutil.parser import parse as pt

def getShortSale(date): #วัน/เดือน/ปี
 context = ssl._create_unverified_context()
 url = 'https://www.set.or.th/set/shortsales.do'
 values = {'from' : date,
          'to' :date,
          'format' : 'html' }

 
 data = parse.urlencode(values).encode()
 req =  request.Request(url, data=data) # this will make the method "POST"
 resp = request.urlopen(req,context=context)
 
 soup = BeautifulSoup(resp.read() , "lxml")
 
 ct = soup.find_all("table",{"class":"table-info"}) 
  
 if(len(ct)>0):
   dt = ct[0].find_all("td")   
   return dt
 else:
   return 0




def shortsale(symbol,dateStart,shorttype="sum"):
  #dateStart="2019-10-09" #10กันยา2019
  symbol = symbol.upper()
  base = datetime.datetime.today()
  x=0
  date_list=[]
  while(True):
    d = base - datetime.timedelta(days=x)
    date_list.append(d)
    x+=1
    if(dateStart==d.strftime("%Y-%m-%d")):
      break   
    
  ############ จบส่วนสร้างช่วงของเวลา
  #symbol = "ADVANC"
  symbolDs = []
  volumeDs = []
  volumnsTotal = []
  dateDs = []
  for i in date_list:
    if(i.weekday()<5):
      #print(calendar.day_name[i.weekday()])
      d=i.strftime("%d/%m/%Y")
      dt = getShortSale(d)
      if(type(dt)!=int):
       k = 0
       for j in range(len(dt)):
        if(symbol in dt[j].text):
         print(".",end="")
         symbolDs.append(dt[j].text)
         volumeDs.append(float(dt[j+1].text.replace(",","")))
         dateDs.append(pt(i.strftime("%Y-%m-%d")))
         k+=1 
         if(k==2):
            break  
  print(" Complete",end="")
  ######################
    
  ###### จัดโครงสร้าง DataFrame
  df = pd.DataFrame({"symbol":symbolDs,"volume":volumeDs,"date":dateDs})
  df = df.set_index("date") 

  rp = df.groupby(df.index)[["volume"]].sum()
  rp.index = rp.index.date  #เอาเวลาออกจาก date คือตัด HH:MM:SS ออก
  return rp



def shortsaleByDay(date,shorttype="sum"):
 date = datetime.datetime.strptime(date,'%Y-%m-%d').strftime("%d/%m/%Y")
   
 context = ssl._create_unverified_context()
 url = 'https://www.set.or.th/set/shortsales.do'
 values = {'from' : date,
          'to' :date,
          'format' : 'html' }

 
 data = parse.urlencode(values).encode()
 req =  request.Request(url, data=data) # this will make the method "POST"
 resp = request.urlopen(req,context=context)
 
 soup = BeautifulSoup(resp.read() , "lxml")
 
 ct = soup.find_all("table",{"class":"table-info"})

 #####
 cr = ct[0].find_all("tr")
 #####
 row = 0
 symbolDs = []
 volumeDs = []   
 for i in cr:
   if(row>0):
      k = i.find_all("td")
      td = 0  
      for j in k:
        if(td==0): 
           symbolDs.append(j.text.replace("-R",""))
        if(td==3): #3 คือ %Short Sale Volume Comparing with Auto Matching
           volumeDs.append(float(j.text.replace(",","").replace("%","")))
           break 
        td+=1
   row+=1    
 
 #######
 df = pd.DataFrame({"symbol":symbolDs,"%ShortSale/AutoMatching":volumeDs})  
 df = df.groupby("symbol")[["%ShortSale/AutoMatching"]].sum()   
 return df


#InterestBearingDebt.py
import requests
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np

def InterestBearingDebt(symbol,keywordlist = ['เงินกู้','ตราสารหนี้','ทรัสต์รีซีท','หุ้นกู้','เงินเบิกเกินบัญชี','หนี้สินระยะยาว - สุทธิจากส่วนที่ถึงกำหนดชำระภายในหนึ่งปี','ส่วนของหนี้สินระยะยาวที่ถึงกำหนดชำระภายในหนึ่งปี'],viewlog=True):
 """
    Returns TL,IBD,Log : หนี้สินรวม , หนี้สินที่คาดว่ามีดอกเบี้ย และ Log ของรายการ 
    
    arg1 (string): ตัวย่อของหุ้น
    
    arg2 (list): keywordlist เป็นตัวคัดกรองข้อมูลส่วน keywordlist เป็นตัวคัดกรองข้อมูล

 """   
###### zone1 ดึงข้อมูล
 headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}  
 symbolMofi = symbol.replace("&","%26")
 #url = "https://www.set.or.th/set/companyhighlight.do?symbol="+symbolMofi
 url = "https://www.set.or.th/set/companyfinance.do?type=balance&symbol="+symbolMofi+"&language=th&country=TH"
 r = requests.get(url,headers)
 soup = BeautifulSoup(r.content, "lxml")

###### zone2 หาส่วนที่จะใช้
 data = soup.find_all("table",{"class":"table table-hover table-info"})
 if(len(data)==0):
    print("ไม่พบข้อมูล")
    return 0,0,0   
 tr = data[2].find_all("tr") 

###### zone3 เตรียมดาต้ามาประมวลผล
 asset = []
 lb = []
 valuelb = []
 
 lb_type = "Current Liabilities"
 lb_type_ds = []

 eq = []
 g = 0
 rp = []  
 previousKey = ""
 for i in tr:
   td = i.find_all("td") 
   if(len(td)==2):
     if(td[1].text=="สินทรัพย์"):
         g+=1
     if(td[1].text=="หนี้สิน"):
         g+=1
     if(td[1].text=="ส่วนของผู้ถือหุ้น"):
         g+=1    
   if(len(td)>2):
    #print(td[1].text)
    #rp.append(td[1].text)
    m = td[2].text.replace(",","")
    m = float(m)   
    #print(m)
    if("\xa0" in td[1].text[0]):
      #print(td[1].text) 
      pass   
 
    if(g==1):
       asset.append(td[1].text)   
    if(g==2):
       
       if(td[1].text=="รวมหนี้สินหมุนเวียน"): #หาว่าหมดสินทรัพย์หมุนเวียนแล้วหรือยัง
          lb_type = "Non-Current Liabilities" 

       if( (td[1].text!="รวมหนี้สินไม่หมุนเวียน") and (td[1].text!="รวมหนี้สินหมุนเวียน")  ):
        
        if(viewlog):
          print(td[1].text+" "+str(m))   
        
        #fix bug 1 may version ของ peter lynch
        previousKey = td[1].text.strip()  
        lb.append(td[1].text)
        valuelb.append(m)

        lb_type_ds.append(lb_type)
        

        '''print(previousKey)
        if("เงินกู้ยืม" in previousKey):
            if(("เงินกู้ยืม" in td[1].text) or ("เงินรับฝาก" in td[1].text)):
              previousKey = "เงินกู้ยืม"
            else:
              previousKey = td[1].text.strip()  
              lb.append(td[1].text)
              valuelb.append(m)
        else:
            previousKey = td[1].text.strip()
            lb.append(td[1].text)
            valuelb.append(m)'''
          
    if(g==3):
        eq.append(td[1].text)        
        
###### zone4 clean พวกแถวที่จะทำให้รายการคำนวณผิด
 mainlabel = []
 valuemainlabel = []
 mainindex = -1
 sublabel = []
 valuesublabel = []
  
 lia_type_main_ds = [] 
 lia_type_sub_ds = []

 #rp1 = []
 for i in range(len(lb)-1): # -1 คือเอา รวมหนี้สิน ออก
  #print(lb[i])
  #rp1.append(lb[i])
  if("\xa0" in lb[i][0]):
    #print(lb[i]) 
    sublabel.append("("+str(mainindex)+") "+lb[i])
    valuesublabel.append(valuelb[i]) 
    lia_type_sub_ds.append(lb_type_ds[i])
  else:
    mainindex+=1
    mainlabel.append("("+str(mainindex)+") "+lb[i])
    valuemainlabel.append(valuelb[i]) 
    lia_type_main_ds.append(lb_type_ds[i])
    
 
###### zone5 คำนวณผลรวม      
 sumlb = 0
 InterestBearingDebt = 0
 #keywordlist = ['เงินกู้','ตราสารหนี้']

 
 rp1 = [] 
 rpv = []  
 rpt = [] 
 for i in range(len(mainlabel)):
  k = 0
  indexSub = 0
  for j in range(len(sublabel)):
    if(mainlabel[i][0:3] in sublabel[j][0:3]):
      k+=1
      subjective = mainlabel[i]+" "+sublabel[j]
      #print(mainlabel[i]+" "+sublabel[j]+ " "+str(valuesublabel[j]))
      #print(subjective)   
      if any(s in subjective for s in keywordlist):    
         #print(mainlabel[i]+" "+sublabel[j]+ " "+str(valuesublabel[j])) 
         rp1.append(mainlabel[i]+" "+sublabel[j])   # + " "+str(valuesublabel[j])
         rpv.append(float(str(valuesublabel[j])))
         InterestBearingDebt += valuesublabel[j]
         rpt.append(lia_type_sub_ds[j])
      #print(valuesublabel[j])         
      sumlb += valuesublabel[j]
  

  if(k==0):
    #print(mainlabel[i])# + " "+str(valuemainlabel[i]))
    subjective = mainlabel[i].strip()#+" "+sublabel[j]
    #print(subjective) 
    if any(s in subjective for s in keywordlist):  
       #print(mainlabel[i]+" "+str(valuemainlabel[i])) 
       rp1.append(mainlabel[i]) #+" "+str(valuemainlabel[i])
       rpv.append(float(str(valuemainlabel[i])))
       InterestBearingDebt += valuemainlabel[i]
       rpt.append(lia_type_main_ds[i])
    #print(valuemainlabel[i])   
    sumlb += valuemainlabel[i]
  
###### zone6 clear report    
 for i in range(len(rp1)): 
  rp1[i] = rp1[i].replace("\xa0","")
  rp1[i] = rp1[i].replace("(1)","")
  rp1[i] = rp1[i].replace("(2)","")
  rp1[i] = rp1[i].replace("(3)","")
  rp1[i] = rp1[i].replace("(4)","")
  rp1[i] = rp1[i].replace("(5)","")
  rp1[i] = rp1[i].replace("(6)","")
  rp1[i] = rp1[i].replace("(7)","")
  rp1[i] = rp1[i].replace("(8)","")
  rp1[i] = rp1[i].replace("(9)","")
  rp1[i] = rp1[i].replace("(10)","")
  rp1[i] = rp1[i].replace("(11)","")
  rp1[i] = rp1[i].replace("(12)","")
  rp1[i] = rp1[i].replace("(0)","")
###### zone7 return ผลลัพธ์
 rp = pd.DataFrame({"Label":rp1,"Value":rpv,"Type":rpt},index=np.array(range(1,len(rp1)+1)))


 ###########
 type_debt = []
 for i in rp["Label"]:
    if("ตราสารหนี้" in i):
       type_debt.append("Bond")
    elif("สถาบันการเงิน" in i):
       type_debt.append("Institute")
    else:
       type_debt.append("N/A")
    
    
 rp["From"] = type_debt

 return sumlb,InterestBearingDebt,rp










try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    #import config as con 
    from starfishX import config as con 
except:
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""

 

from sklearn.ensemble import IsolationForest
from sklearn import svm
from sklearn.covariance import EllipticEnvelope
#LocalOutlierFactor
from sklearn.neighbors import LocalOutlierFactor

from sklearn.preprocessing import StandardScaler

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum



from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

class anomalyDetectionAgloType(Enum):
 IsolationForests = "IsolationForests"
 OneClassSVM   = "OneClassSVM"
 GaussianDistribution = "GaussianDistribution"
 LocalOutlierFactor = "LocalOutlierFactor"

class anomalyDetectionDataType(Enum):
 priceChg = "priceChg"
 volume = "Volume"

def AnomalyDetection(dataset,contamination=0.01,Lastday=5,dataType="",algoType="",viewplot=False,viewlog=False,padding=False):
   ###############
   try:
    if(len(dataset.columns)!=2):
     print("dataset Not Match")
     return 0,0
   except:
     print("dataset Not Match")
     return 0,0

   outliers_fraction = contamination
   
   if(dataType==""):
      dataType = anomalyDetectionDataType.volume

   if(algoType==""):
      algoType = anomalyDetectionAgloType.IsolationForests

   if(not isinstance(dataType, anomalyDetectionDataType)):
    print("dataType Not Match")
    return 0,0

   if(not isinstance(algoType, anomalyDetectionAgloType)):
    print("algorithm Type Not Match")
    return 0,0
    
   alType = algoType.value
   dType = dataType.value
   
   if(dType=="priceChg"):
     dataset = dataset[[dataset.columns[0]]]
     dataset = dataset[[dataset.columns[0]]].pct_change().dropna()
   elif(dType=="Volume"):
     dataset = dataset[[dataset.columns[1]]]
     if(not any("Volume" in s for s in dataset.columns)):
      print("DataSet Not Contain Volume")
      return 0,0

   if(alType == "LocalOutlierFactor"):
     #LocalOutlierFactor
     # fit the model for novelty detection (novelty=True)
     n_neighbors = int(len(dataset)*.1)
     clf = LocalOutlierFactor(algorithm="auto",n_neighbors=n_neighbors, novelty=True, contamination=outliers_fraction)
     clf.fit(dataset)
     # DO NOT use predict, decision_function and score_samples on X_train as this
     # would give wrong results but only on new unseen data (not used in X_train),
     # e.g. X_test, X_outliers or the meshgrid
     #y_pred_test = clf.predict(X_test)
     dataset['anomaly'] = clf.predict(dataset)
     labelTitle = "(LocalOutlierFactor For Anomaly Detection)"

   if(alType == "IsolationForests"):
     rng = np.random.RandomState(42)
     data = dataset
     scaler = StandardScaler()
     np_scaled = scaler.fit_transform(data)
     data = pd.DataFrame(np_scaled)
     # train isolation forest
     model =  IsolationForest(behaviour='new',contamination=outliers_fraction,
                      random_state=rng)
     model.fit(data) 
     dataset['anomaly'] = model.predict(data)
     labelTitle = "(Isolation Forests For Anomaly Detection)"
   
   if(alType == "OneClassSVM"):
     data = dataset
     scaler = StandardScaler()
     np_scaled = scaler.fit_transform(data)
     data = pd.DataFrame(np_scaled)
     # train oneclassSVM 
     model = svm.OneClassSVM(nu=outliers_fraction, kernel="rbf", gamma=0.01)
     model.fit(data)
     dataset['anomaly'] =  model.predict(data)
     labelTitle = "(OneClassSVM For Anomaly Detection)"
     if(viewlog==True):
       print("OneClassSVM/rbf/gamma 0.01")

   if(alType == "GaussianDistribution"):
     envelope =  EllipticEnvelope(contamination = outliers_fraction,random_state=42) 
     X_train = dataset.values.reshape(-1,1)
     cov = envelope.fit(X_train)
      
     #k1 = cov.decision_function(X_train)
     dataset["anomaly"] = cov.predict(X_train)
     labelTitle = "(GaussianDistribution For Anomaly Detection)"

   if(viewplot==True): 
    symbol = dataset.columns[0].split(":")[0]
    fig = plt.figure(figsize=(16,8))
    fig.subplots_adjust(hspace=0.4, wspace=0.4)

    if(padding==False):
      ax = fig.add_subplot(1, 1, 1) #แถว #หลัก #รูปที่
      ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
      ax = dataset[dataset.columns[0]].plot(figsize=(20,8))
      dataset[dataset.columns[0]][(dataset['anomaly']==-1)].plot(ax=ax,marker="*",ms=10,color='r',linewidth=0)
      return 0,0

    elif(padding==True):
      fig, ax = plt.subplots(figsize=(20,8)) #แถว #หลัก #รูปที่
      ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
      ax.plot(dataset[dataset.columns[0]]) #figsize=(30,8),title=title
      ax.plot(dataset[dataset.columns[0]][(dataset['anomaly']==-1)],"*",ms=10,color="r",linewidth=0)
      plt.margins(0.01,0.05)  
      return 0,0


   #ตรวจ จำนวนวันหลังสุด พบสิ่งผิดปกติไหม
   result = list(map(lambda x: x==-1, dataset.tail(Lastday)['anomaly']))
   if(True in result):
      return dataset.columns[0].split(":")[0],dataset
   else:
      return "Normal",dataset



def viewAnomalyDetectionReportAllType(dataset,dataType,contamination=0.01,padding=False):
 fig = plt.figure(figsize=(24,12))
 fig.subplots_adjust(hspace=0.4, wspace=0.1)
 symbol = dataset.columns[0].split(":")[0]
 
 dType = dataType.value
 outliers_fraction = contamination

 ###############
 ax = fig.add_subplot(2, 2, 1)
 
 kAlgo = sx.anomalyDetectionAgloType.LocalOutlierFactor
 _,rp = sx.AnomalyDetection(dataset,dataType=dataType,algoType=kAlgo
                          ,Lastday=5,contamination=outliers_fraction,viewplot=False)
 
 labelTitle = "("+kAlgo.value+" For Anomaly Detection)"

 if(padding==False):                         
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax = rp[rp.columns[0]].plot(figsize=(24,12))
   rp[rp.columns[0]][(rp['anomaly']==-1)].plot(ax=ax,marker="*",ms=10,color='r',linewidth=0)
 
 if(padding==True):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax.plot(rp[rp.columns[0]]) #figsize=(30,8),title=title
   ax.plot(rp[rp.columns[0]][(rp['anomaly']==-1)],"*",ms=10,color="r",linewidth=0)
   plt.margins(0.01,0.05)

 ###############
 ax = fig.add_subplot(2, 2, 2)

 kAlgo = sx.anomalyDetectionAgloType.IsolationForests
 _,rp = sx.AnomalyDetection(dataset,dataType=dataType,algoType=kAlgo
                          ,Lastday=5,contamination=outliers_fraction,viewplot=False)
 labelTitle = "("+kAlgo.value+" For Anomaly Detection)"

 if(padding==False):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax = dataset[rp.columns[0]].plot()
   rp[rp.columns[0]][(rp['anomaly']==-1)].plot(ax=ax,marker="*",ms=10,color='r',linewidth=0)

 if(padding==True):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax.plot(rp[rp.columns[0]]) #figsize=(30,8),title=title
   ax.plot(rp[rp.columns[0]][(rp['anomaly']==-1)],"*",ms=10,color="r",linewidth=0)
   plt.margins(0.01,0.05)
 ###############

 ###############
 ax = fig.add_subplot(2, 2, 3)

 kAlgo = sx.anomalyDetectionAgloType.GaussianDistribution
 _,rp = sx.AnomalyDetection(dataset,dataType=dataType,algoType=kAlgo
                          ,Lastday=5,contamination=outliers_fraction,viewplot=False)
 labelTitle = "("+kAlgo.value+" For Anomaly Detection)"

 if(padding==False):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax = rp[rp.columns[0]].plot()
   rp[rp.columns[0]][(rp['anomaly']==-1)].plot(ax=ax,marker="*",ms=10,color='r',linewidth=0)

 if(padding==True):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax.plot(rp[rp.columns[0]]) #figsize=(30,8),title=title
   ax.plot(rp[rp.columns[0]][(rp['anomaly']==-1)],"*",ms=10,color="r",linewidth=0)
   plt.margins(0.01,0.05)
 ###############

 ###############
 ax = fig.add_subplot(2, 2, 4)

 kAlgo = sx.anomalyDetectionAgloType.OneClassSVM
 _,rp = sx.AnomalyDetection(dataset,dataType=dataType,algoType=kAlgo
                          ,Lastday=5,contamination=outliers_fraction,viewplot=False)
 labelTitle = "("+kAlgo.value+" For Anomaly Detection)"

 if(padding==False):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax = rp[rp.columns[0]].plot()
   rp[rp.columns[0]][(rp['anomaly']==-1)].plot(ax=ax,marker="*",ms=10,color='r',linewidth=0)

 if(padding==True):
   ax.title.set_text(symbol.upper()+' : '+dType+" "+labelTitle)
   ax.plot(rp[rp.columns[0]]) #figsize=(30,8),title=title
   ax.plot(rp[rp.columns[0]][(rp['anomaly']==-1)],"*",ms=10,color="r",linewidth=0)
   plt.margins(0.01,0.05)
 ###############


######### moc.py
try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    from starfishX import config as con 
    #print("Dev")
except:
    #print("Debug")
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""


import pandas as pd
import requests
import numpy as np
from bs4 import BeautifulSoup
import json
import matplotlib.pyplot as plt
 
#%config InlineBackend.figure_format ='retina'
import matplotlib as mpl
import matplotlib.font_manager as font_manager

import matplotlib.dates as mdates
from datetime import datetime




def mocExport(keyword,year,month,notkeyword="xyzabc0123456789"):

   #
   #set title ภาษาไทย
   #https://stackoverflow.com/questions/16574898/how-to-load-ttf-file-in-matplotlib-using-mpl-rcparams
   path = 'DB_Helvethaica_X.ttf'
   prop = font_manager.FontProperties(fname=path)
   mpl.rcParams['font.family'] = prop.get_name()
 
   params = {'legend.title_fontsize':0,
          'legend.fontsize': 20,
          'figure.figsize': (16, 8),
          'axes.labelsize': 20,
          'axes.titlesize':20,
          'xtick.labelsize':20,
          'ytick.labelsize':20}
   plt.rcParams.update(params)


   #keyword = "อัญมณี"
   year = str(year)
   #month = ["7","8","9"]

   url = "https://dataapi.moc.go.th/products?keyword="+keyword+"&revision=2017&imex_type=export&order_by=com_code"
   headers = {"User-Agent": "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"}
  
   r = requests.get(url,headers)
   jsonContent = json.loads(r.content)
   print("Cat Item : "+str(len(jsonContent)).strip())
   if(len(jsonContent)>30):
     print("Keyword มากเกินไป เพิ่มคำระบุในการค้นหา")
     return 0,0
   
   if(len(jsonContent)==0):
     print("ไม่พบข้อมูล")
     return 0,0
   ###################  # hs_description_th com_description_th
   keyCodeDS = pd.DataFrame(jsonContent).groupby("com_code")[["com_code","com_description_th","hs_description_th"]].max()
   
   content_ds = []
   for i,col in enumerate(keyCodeDS["com_code"],0):
    com_code = keyCodeDS["com_code"][i]
    print(keyCodeDS["com_code"][i],keyCodeDS["com_description_th"][i].strip())
    for m in month:
     url = "https://dataapi.moc.go.th/export-commodity-countries?year="+ \
        year+"&month="+str(m)+"&com_code="+keyCodeDS["com_code"][i]+"&limit="
     r = requests.get(url,headers)
     itemContent = json.loads(r.content) 
     df = pd.DataFrame(itemContent)
     if(len(df)>0):
       value_item = df[["country_name_th","value_baht"]]["value_baht"].sum()
     else:
       value_item = 0
       
     value_item = value_item/1000000   
   
     #print(year,m)
     month_ds = datetime(int(year),int(m),1)
     desc = keyCodeDS["com_description_th"][i].strip().replace(" ","")
     
     try:
      descDetail = keyCodeDS["hs_description_th"][i].strip().replace(" ","")
     except:
      descDetail = "None"

     #ตรวจสอบเช่นตาไก่ จะเป็นอุปกรณ์ไม่ใช่สินค้าเกษตร
     if(not((notkeyword in descDetail) or (notkeyword in desc)) ):
        content_ds.append({"Month":month_ds,"Label":desc,"Value":value_item,"Detail":descDetail})
        
   #######################
   k = pd.DataFrame(content_ds) 
    
   pivot_df = k.pivot(index='Month', columns='Label', values='Value') 
   labels = [l.strftime('%Y-%m') for l in pivot_df.index]
    
   ax = pivot_df.plot(figsize=(10,6),kind="bar",stacked=True,title="ยอดส่งออกสินค้า")
   ax.set_xticklabels(labels)
   ax.set_ylabel("ล้านบาท")
   plt.legend(loc='upper left', bbox_to_anchor=(1.0, 1.0))
   plt.rc('font', family='serif', size='12.0')
   

   return k,pivot_df


from starfishX.markminervini.trendtemplate import trendMA_YMF
from starfishX.markminervini.trendtemplate import trendMA_Year
from starfishX.markminervini.trendtemplate import check52WeekHighLow
from starfishX.markminervini.trendtemplate import ROC_Basket

__version__ = "markminervini 0.1" #2022-oct-31


try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    from starfishX import config as con 
    #print("Dev")
except:
    #print("Debug")
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""


exec("from "+prefix+"starfishXfn import setTHforPlot")

#from starfishX.starfishXfn import loadHistData
exec("from "+prefix+"starfishXfn import loadHistData")

#from starfishX.starfishXfn import MonthlyReturn
exec("from "+prefix+"starfishXfn import MonthlyReturn")

#from starfishX.starfishXfn import listShareholders
exec("from "+prefix+"starfishXfn import listShareholders")

#from starfishX.starfishXfn import findCASH
exec("from "+prefix+"starfishXfn import findCASH")

#from starfishX.starfishXfn import commonsizeBS
exec("from "+prefix+"starfishXfn import commonsizeBS")

#from starfishX.starfishXfn import commonsizeIS
exec("from "+prefix+"starfishXfn import commonsizeIS")

#from starfishX.starfishXfn import getFinanceRatio
exec("from "+prefix+"starfishXfn import getFinanceRatio")

#from starfishX.starfishXfn import spiderCompare
exec("from "+prefix+"starfishXfn import spiderCompare")

#from starfishX.starfishXfn import getBalanceSheet
exec("from "+prefix+"starfishXfn import getBalanceSheet")

#from starfishX.starfishXfn import getBalanceSheetCompare
exec("from "+prefix+"starfishXfn import getBalanceSheetCompare")

#exec("from "+prefix+"starfishXfn import getProfitAbility")  #ยกเลิกก่อน  เพราะ wsj หยุดให้บริการแบบ free
exec("from "+prefix+"getProfitAbility import getProfitAbility")  #ยกเลิกก่อน  เพราะ wsj หยุดให้บริการแบบ free

#from starfishX.starfishXfn import getMemberOfIndex
exec("from "+prefix+"starfishXfn import getMemberOfIndex")

#from starfishX.starfishXfn import getMemberOfIndex
exec("from "+prefix+"starfishXfn import getDividendYieldOfIndex")

#from starfishX.starfishXfn import listStockInSector
# exec("from "+prefix+"starfishXfn import listStockInSector")

#from starfishX.starfishXfn import getFundamentalInSector
#exec("from "+prefix+"starfishXfn import getFundamentalInSector")   fixbug after settrade change webpage
#version > 0.155538
exec("from "+prefix+"getFundamentalInSector import getFundamentalInSector") 
exec("from "+prefix+"getFundamentalInSector import listStockInSector")
exec("from "+prefix+"getFundamentalInSector import getPriceAndRatioInSector") 
exec("from "+prefix+"getFundamentalInSector import getRatioAndBalanceSheet") 
#from starfishXfn import evperebitda #ยกเลิกก่อน

#from starfishX.starfishXfn import MarkowitzProcess
exec("from "+prefix+"starfishXfn import MarkowitzProcess")

#from starfishX.listSecurities import listSecurities
exec("from "+prefix+"listSecurities import listSecurities")

#from starfishX.findMarketCap import findMarketCap
exec("from "+prefix+"findMarketCap import findMarketCap")

#from starfishX.incomestatement import getIncomeStatement
exec("from "+prefix+"incomestatement import getIncomeStatement")

#from starfishX.incomestatement import getIncomeStatementCompare
exec("from "+prefix+"incomestatement import getIncomeStatementCompare")

#from starfishX.DuPontROE import DuPontROE
exec("from "+prefix+"DuPontROE import DuPontROE")

#from starfishX.InterestBearingDebt import InterestBearingDebt
exec("from "+prefix+"InterestBearingDebt import InterestBearingDebt")

#from starfishX.depreciationAndAmortisation import DA
exec("from "+prefix+"depreciationAndAmortisation import DA")

#from starfishX.getEVEBITDA import getEVEBITDA
exec("from "+prefix+"getEVEBITDA import getEVEBITDA")

#from starfishX.getEVEBITDA import getEVEBITDA
exec("from "+prefix+"getEVEBITDA import getEVEBIT")

#from starfishX.getEVEBITDA import analyseEVEBITDA
exec("from "+prefix+"getEVEBITDA import analyseEVEBITDA")

#from starfishX.morningstarGetfn import MStar
exec("from "+prefix+"morningstarGetfn import MStar")

#from starfishX.morningstarGetfn import morningstarGetfn
exec("from "+prefix+"morningstarGetfn import morningstarGetfn")

#from starfishX.DividendDiscountModel import DividendDiscountModel
exec("from "+prefix+"DividendDiscountModel import DividendDiscountModel")

#from starfishX.Ke import governmentBond10year
exec("from "+prefix+"Ke import governmentBond10year")

#from starfishX.Ke import beta
exec("from "+prefix+"Ke import beta")

#from starfishX.checkServices import checkServices
exec("from "+prefix+"checkServices import checkServices")

#from starfishX.indicator import linregressIndicator
exec("from "+prefix+"indicator import linregressIndicator")

#from starfishX.indexMarket import indexMarket
exec("from "+prefix+"indexMarket import indexMarket")

#from starfishX.marketview  import marketview
exec("from "+prefix+"marketview import marketview")

#from starfishX.marketview  import marketview
exec("from "+prefix+"marketview import marketViewForeignTrade")

#from starfishX.AD  import AnomalyDetectionVolume
exec("from "+prefix+"AD import AnomalyDetection")

#from starfishX.AD  import AnomalyDetectionVolume
exec("from "+prefix+"AD import anomalyDetectionAgloType")

#from starfishX.AD  import AnomalyDetectionVolume
exec("from "+prefix+"AD import anomalyDetectionDataType")

#from starfishX.AD  import AnomalyDetectionVolume
exec("from "+prefix+"AD import viewAnomalyDetectionReportAllType")

#from starfishX.Ke import governmentBond10year
exec("from "+prefix+"Ke import thaibma")
exec("from "+prefix+"Ke import thaibma_series")
exec("from "+prefix+"Ke import BondType") #enum bondType

exec("from "+prefix+"EYG import EarningYieldGap")

exec("from "+prefix+"EYG import twoTenSpread")
exec("from "+prefix+"EYG import utilityTwinXPlot")


exec("from "+prefix+"starfishXfn import RelativeStrengthRank")
exec("from "+prefix+"starfishXfn import RelativeStrength_hist")
exec("from "+prefix+"starfishXfn import RelativeStrength_hist_with_progress1")
exec("from "+prefix+"starfishXfn import RelativeStrength_hist_with_progress2")

exec("from "+prefix+"Elder import elderSafezone")


exec("from "+prefix+"openInterest import listSSF")   
exec("from "+prefix+"openInterest import OpenInterestContracts")

exec("from "+prefix+"moc import mocExport")

exec("from "+prefix+"marketValueDaily import marketValueDaily")

exec("from "+prefix+"marketValueDaily import getTotalShare")

exec("from "+prefix+"shortsale import shortsale")
exec("from "+prefix+"shortsale import shortsaleByDay")


#from starfishX.starfishXfn import loadHistData
exec("from "+prefix+"starfishXfn import fillHistData")

exec("from "+prefix+"utility import *")

exec("from "+prefix+"risk_kit import *")


exec("from "+prefix+"nvdr import getNVDRVolume")
exec("from "+prefix+"nvdr import reduceXtickfrequency") 

exec("from "+prefix+"news_api import news_api_bydate")
exec("from "+prefix+"news_api import news_api")
exec("from "+prefix+"news_api import NewsTypeMatch")

#exec("from "+prefix+"sentimentAnalysis import sentimentAnalysis")


exec("from "+prefix+"ShareholdersNow import ShareholdersNow")
exec("from "+prefix+"ShareholdersNow import ShareholdersHist")

exec("from "+prefix+"holidayMarket import getHolidayMarket")

exec("from "+prefix+"ipoStat import ipoStat")


#from starfishX.starfishXfn import loadHistData
exec("from "+prefix+"loadHistData_v2 import loadHistData_v2")
exec("from "+prefix+"loadHistData_v2 import timeFrame")

exec("from "+prefix+"getWarrant import getWarrant")
exec("from "+prefix+"getDW import getDW")

exec("from "+prefix+"greek import days_between")
exec("from "+prefix+"greek import EuropeanCall")
exec("from "+prefix+"greek import EuropeanCallRolling")

#exec("from "+prefix+"utilHelpers import rankWithRange")
#exec("from "+prefix+"utilHelpers import HistStockPrice")
 

__version__ = 0.155550  #21 Feb 2023 update 0.155549

__doc__ = """starfishX Version จัดเตรียมข้อมูลสำหรับประมวลผลการลงทุนในตลาดหุ้นไทย """


import os
import pandas as pd
import datetime

from enum import Enum

from IPython.display import HTML

import re
english_check = re.compile(r'[a-zA-Z]')

class NewsTypeMatch(Enum):
 any = "any"
 all = "all"


from datetime import date, timedelta

def daterange(start_date, end_date):
    for n in range(int((end_date - start_date).days)):
        yield start_date + timedelta(n)


def news_api_bydate(start_date,end_date,keyword):
    m1 = start_date.split('-')
    m2 = end_date.split('-')
    
    start_date = date(int(m1[0]), int(m1[1]), int(m1[2]))
    end_date = date(int(m2[0]), int(m2[1]), int(m2[2])) + timedelta(1)
    
    datecontent = [] 
    content_news = []
    for single_date in daterange(start_date, end_date):
      dateRange = str(single_date.strftime("%Y-%m-%d"))
     
      #ดึงข้อมูลใน Directory
      arr = os.listdir("news")
      #keyword = symbol

      dateDS = []
      content = []
      for filename in arr:
        file = "news/"+filename
        dateFilename = filename.replace("news-","").replace(".txt","")[:-3]
        
        #print(type(dateFilename),type(dateRange),dateRange,dateFilename)
        if(dateFilename==dateRange): 
          with open(file) as f:
            for line in f:
              lineP = line.lower() 
             
              for k in keyword:
                if(k in lineP):
                  #print(dateFilename+lineP)
                  datecontent.append(dateFilename)
                  lineP = lineP.replace('\n',' ')
                  content_news.append(lineP)  
                
    pd.set_option('display.max_colwidth', None)            
    return pd.DataFrame({'date':datecontent,'content':content_news})

def news_api(keyword,contentLen=100,contentShort=True,typeMatch=NewsTypeMatch.any,highlight=True):
 '''
 keyword : (list) คำที่ต้องการค้นหาในข่าว
 typeMatch : (str) มีสอง type คือ any คือคำใดคำหนึ่ง และ all คือต้อง match ทั้ง list
 '''   
 
 if(type(keyword)!=list):
   print("keyword must Type List")
   return 0

 #ดึงข้อมูลใน Directory
 arr = os.listdir("news")
 #keyword = symbol

 dateDS = []
 content = []
 for filename in arr:
  file = "news/"+filename
  date = filename.replace("news-","").replace(".txt","")[:-3]
  
  if('DS_Store' in file):
       continue 
      
  with open(file) as f:
   for line in f:
    lineP = line.lower() 
    #matching = [s for s in keyword if s.lower() in lineP]
    matching = []
    for s in keyword:
      if(s.lower() in lineP):
         #print(s.lower(),lineP)
         if(english_check.match(s.lower())):
           if(highlight):
             #print(s.upper(),s.lower())
             line = line.replace(s.upper(),"<mark>"+s.upper()+"</mark>")
             line = line.replace(s.lower(),"<mark>"+s.lower()+"</mark>")
             line = line.replace(s,"<mark>"+s+"</mark>")  #เพิ่มกรณี ตัวอักษรตัวแรกเป็นตัวใหญ่ เช่น Spin-off
           matching.append(s)
         else:
           if(highlight):
             line = line.replace(s,"<mark>"+s+"</mark>")
           matching.append(s)
    #  else:
    #     matching.append(None)
    #print(matching)


    if(typeMatch.value=="any"):
     if(len(matching)>0):
      dateDS.append(datetime.datetime.strptime(date, '%Y-%m-%d'))
      content.append(line)
    
    if(typeMatch.value=="all"):
     if(len(matching)==len(keyword)):
      dateDS.append(datetime.datetime.strptime(date, '%Y-%m-%d'))
      content.append(line)

 ################    
 if(len(content)==0):
    return 0

 dfNews = pd.DataFrame({"Date":dateDS,"content":content})

 #######################
 #ยุบข่าวให้ไม่ยาวเกินไป
 tmp = dfNews["content"].str.split(" ")
 shortContent = []
 for i in tmp:
  s = ""
  for j in i:
   if(len(s)+len(j)<=contentLen):
     s+=j+" "
     #print(j)
  
  shortContent.append(s.replace("\n","")) 

 pd.set_option('display.max_colwidth', None) # -1
 if(contentShort):
     dfNews["contentShort"] = shortContent
     df = dfNews[["Date","contentShort"]].sort_values(["Date"],ascending=False)
     if(highlight):
      df = HTML("<style> mark {background-color: yellow;color: black;} </style>"+df.to_html(escape=False))
      return df# = HTML(df.to_html(escape=False))

     return df 
 else:
     if(highlight):
       df = HTML("<style> mark {background-color: yellow;color: black;} </style>"+df.to_html(escape=False))
       return df# = HTML(df.to_html(escape=False))
     return dfNews.sort_values(["Date"],ascending=False)


### EYG.py
try:
    #ลำดับมีผลพยายามโหลดที่ ใน folder ที่กำลัง Implement ก่อน
    from starfishX import config as con 
    #print("Dev")
except:
    #print("Debug")
    import starfishX.config as con

prefix = "starfishX."

if(con.__mode__=="debug"):
    prefix = ""

 
 
import numpy as np

import matplotlib.dates as mdates

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

import matplotlib.pyplot as plt
import pandas as pd

exec("from "+prefix+"marketview import marketview")
exec("from "+prefix+"Ke import thaibma")

def twoTenSpread(start="2000-01-01",bondtype="",viewplot=True,padding=False):
 
 ####### เตรียมข้อมูล #########
 date = datetime(2000,1,1)#startdate
 bt = bondtype
 stop = datetime.now()

 k2ds = []
 k10ds = []
 dateIndex = []
 for i in range(1000): #เดาว่าไม่เกิน 1000 เดือน
  dt_str = date.strftime('%Y-%m-%d')
 
  k2_10 = thaibma(date=dt_str,bondtype=bt,viewlog=False)

  if(k2_10==0): #พยายามเลื่อนวันถัดไป ถ้าดึงข้อมูลไม่ได้
    for dNext in range(10): 
      dt_str = (date + pd.DateOffset(dNext)).strftime('%Y-%m-%d')
      k2_10 = thaibma(date=dt_str,bondtype=bt,viewlog=False)    
      if(k2_10!=0):
        break
 
  dateIndex.append(dt_str)
  k2ds.append(k2_10["USTreasury2Year"])
  k10ds.append(k2_10["USTreasury10Year"]) 
       
  if((stop.month-1==date.month) & (stop.year==date.year)):
   break

  date = date + relativedelta(months=+1)
    
    
 ############## 
 #https://stackoverflow.com/questions/45704366/modify-date-ticks-for-pandas-plot
 # convert date objects from pandas format to python datetime
 index = pd.date_range(start = "2000-01-01", end = "2019-06-01", freq = "MS")
 dateIndex = [pd.to_datetime(date, format='%Y-%m-%d').date() for date in index]
 df = pd.DataFrame({"USTreasury2Year":k2ds,"USTreasury10Year":k10ds},index=dateIndex)    
 df["2-10Spread"] = df["USTreasury10Year"] - df["USTreasury2Year"]

 ####### จบส่วนเตรียมข้อมูล #######   
    
 fig = plt.figure(figsize=(20,10))
 fig.subplots_adjust(hspace=0.4, wspace=0.1)
 ax = fig.add_subplot(1, 1, 1)
 ax.plot(df["2-10Spread"])  

 ax.title.set_text("2-10 Spread : (USTreasury10Year - USTreasury2Year)")
 ax.xaxis.set_major_locator(mdates.MonthLocator(interval=12))
 ax.xaxis.set_minor_locator(mdates.MonthLocator(interval=1))
 # set formatter
 ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))
 #ax.xaxis.set_minor_formatter(mdates.DateFormatter('%m'))
 plt.grid()
 plt.gcf().autofmt_xdate()
 plt.margins(0.01,0.05)    
    
 return df


def utilityTwinXPlot(Left,Right,LeftYLim,RightYLim,title):
 fig, ax1 = plt.subplots(figsize=(20,10))
 ax1.title.set_text(title)

 ax1.plot(Left.index,Left,color='red',label=Left.columns[0]+"(LHS)")
 ax1.set_ylim(LeftYLim)

 ax1.legend(loc='upper right', bbox_to_anchor=(1,1))

 ax1.grid()
 ax1.xaxis.set_major_locator(mdates.MonthLocator(interval=12))
 ax1.xaxis.set_minor_locator(mdates.MonthLocator(interval=1))
 # set formatter
 ax1.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))
 #ax.xaxis.set_minor_formatter(mdates.DateFormatter('%m'))

 plt.gcf().autofmt_xdate()
 plt.margins(0.01,0.05)


 ax2 = ax1.twinx()
 ax2.plot(Left.index,Right,color="blue",label=Right.columns[0]+"(RHS)")
 ax2.set_ylim(RightYLim)
 ax2.legend(loc='upper right')

 ax2.xaxis.set_major_locator(mdates.MonthLocator(interval=12))
 ax2.xaxis.set_minor_locator(mdates.MonthLocator(interval=1))
 # set formatter
 ax2.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))
 #ax.xaxis.set_minor_formatter(mdates.DateFormatter('%m'))
 ax2.legend(loc='upper right', bbox_to_anchor=(1,0.9))
 plt.grid()
 plt.gcf().autofmt_xdate()
 plt.margins(0.01,0.05)

 fig.tight_layout()  # otherwise the right y-label is slightly clipped
 plt.show()   


def EarningYieldGap(marketType,BondType,start,viewlog=True,viewplot=True):
 '''
 ตัวอย่างการใช้พารามิเตอร์
 #marketType = sx.indexMarket.SET50
 #BondType = sx.BondType.USTreasury10Year
 '''

 marketDictKey = marketType.value
 df = marketview(marketType,viewplot=False,start=start)
 ep = "E/P "+marketDictKey
 df[ep] = 1/df["pe"]
 bondYield = []
 for i in df.index:
  dt = i.strftime('%Y-%m-%d') #จัดรูปแบบ Date ใหม่  คือตัดช่วงของเวลาออก 00:00:00
  by = thaibma(dt,bondtype=BondType,viewlog=False)
  if(by==0): #พยายามเลื่อนวันถัดไป ถ้าดึงข้อมูลไม่ได้
    for k in range(10): 
      dt = (i + pd.DateOffset(k)).strftime('%Y-%m-%d')
      by = thaibma(dt,bondtype=BondType,viewlog=False)    
      if(by!=0):
        break
  
  if(viewlog==True):
    print(".",end="")   
  bondYield.append(by)

 if(viewlog==True):
   print("complete")

 typeBondKey = "BondYield "+BondType.value
 df[typeBondKey] = bondYield
 ######## end process data #######
 
 ###### part plot ########   
 if(viewplot==True):
  fig = plt.figure(figsize=(20, 20))
  grid = plt.GridSpec(4, 1, wspace=0.4, hspace=0.3)
  ax = plt.subplot(grid[0, 0])
  df["index"].plot(ax=ax,title=marketDictKey+" Index")

  ax = plt.subplot(grid[1, 0])
  df["pe"].plot(ax=ax,title=df.columns[1].upper()+" "+marketDictKey+ " Index")

  ax = plt.subplot(grid[2:, 0])
  df[[ep,typeBondKey]].plot(ax=ax,
                kind="line",style='--',title="Earning Yield Gap ("+marketDictKey+" & "+typeBondKey+")")

 
 ######### end part plot #######
 return df               